# !/usr/bin/env python3
# Copyright (c) 2024 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Callable, List, Optional, Union

import numpy as np
import torch

from quairkit.core import utils
from quairkit.core.intrinsic import _is_sample_linear

from ..core import utils
from ..core.intrinsic import _ArrayLike, _StateLike, _type_transform

__all__ = [
    "is_choi",
    "is_density_matrix",
    "is_hermitian",
    "is_linear",
    "is_positive",
    "is_povm",
    "is_projector",
    "is_ppt",
    "is_pvm",
    "is_state_vector",
    "is_unitary",
]


def is_choi(op: _ArrayLike) -> Union[bool, List[bool]]:
    r"""Check if the input op is a Choi operator of a physical operation.
        Support batch input.

    Args:
        op: matrix form of the linear operation.

    Returns:
        Whether the input op is a valid quantum operation Choi operator. 
        For batch input, return a boolean array with the same batch dimensions as input.

    Note:
        The operation op is (default) applied to the second system.

    .. code-block:: python

        ischoi = is_choi(eye(4) / 4)
        print(f'The operation is choi or not:\n{ischoi}')

    ::

        The operation is choi or not:
        True

    """
    op = _type_transform(op, "tensor")

    assert utils.check._is_square(op), \
        f"The input matrix is not a square matrix: received shape {op.shape}"

    return utils.check._is_choi(op).tolist()


def is_density_matrix(
    rho: _ArrayLike, eps: float = 1e-6
) -> Union[bool, List[bool]]:
    r"""Verify whether ``rho`` is a legal quantum density matrix
        Support batch input

    Args:
        rho: density matrix candidate
        eps: tolerance of error, default to be `None` i.e. no testing for data correctness

    Returns:
        determine whether ``rho`` is a PSD matrix with trace 1


    .. code-block:: python

        is_dens_mat = is_density_matrix(eye(4) / 4)
        print(f'The operation is density matrix or not:\n{is_dens_mat}')

    ::

        The operation is density matrix or not:
        True

    """
    rho = _type_transform(rho, "tensor")

    assert utils.check._is_square(rho), \
        f"The input matrix is not a square matrix: received shape {rho.shape}"

    return utils.check._is_density_matrix(rho, eps).tolist()


def is_hermitian(
    mat: _ArrayLike, eps: float = 1e-6
) -> Union[bool, List[bool]]:
    r"""Verify whether ``mat`` is Hermitian.
        Support batch input.

    Args:
        mat: hermitian candidate :math:`P`
        eps: tolerance of error

    Returns:
        determine whether :math:`P - P^\dagger = 0`
        For batch input, return a boolean array with the same batch dimensions as input.

    .. code-block:: python

        is_hermit = is_hermitian(x())
        print(f'The operation is hermitian or not:\n{is_hermit}')

    ::

        The operation is hermitian or not:
        True

    """
    mat = _type_transform(mat, "tensor")

    assert utils.check._is_square(mat), \
        f"The input matrix is not a square matrix: received shape {mat.shape}"

    return utils.check._is_hermitian(mat, eps).tolist()


def is_linear(
    func: Callable[[Union[torch.Tensor, np.ndarray]], Union[torch.Tensor, np.ndarray]],
    info: Union[List[int], Callable[[], np.ndarray], Callable[[], torch.Tensor]],
    input_dtype: torch.dtype = None,
    eps: Optional[float] = 1e-5,
) -> bool:
    r"""Check if the provided function 'func' is linear.

    Args:
        func: A callable function to be tested. This function should accept and return either a torch.Tensor or a numpy.ndarray.
        info: A parameter specifying the shape of the input for 'func'. It can be a list of two integers (defining the shape of a tensor),
              a callable that returns a numpy.ndarray, or a callable that returns a torch.Tensor.
        eps: An optional tolerance value used to determine if the function's behavior is close enough to linear. Default value is 1e-6.

    Returns:
        bool: True if 'func' behaves as a linear function within the specified tolerance 'eps'; False otherwise.

    Raises:
        TypeError: If 'func' is not callable, does not accept a torch.Tensor or numpy.ndarray as input, or does not return a torch.Tensor or numpy.ndarray.
                   If 'info' is not a valid type (not a list of integers or a callable returning a torch.Tensor or numpy.ndarray).

    .. code-block:: python

       def f(X):
            return torch.trace(X)

        is_lin = is_linear(f, info=[2, 2])
        print(f'The function f is linear or not:\n{is_lin}')

    ::

        The function f is linear or not:
        True

    """
    func, generator, input_dtype, _ = _is_sample_linear(func, info, input_dtype)
    return utils.check._is_linear(
        func=func, generator=generator, input_dtype=input_dtype, eps=eps
    )


def is_positive(
    mat: _ArrayLike, eps: float = 1e-6
) -> Union[bool, List[bool]]:
    r"""Verify whether ``mat`` is a positive semi-definite matrix.
        Support batch input.

    Args:
        mat: positive operator candidate :math:`P`
        eps: tolerance of error

    Returns:
        determine whether :math:`P` is Hermitian and eigenvalues are non-negative
        For batch input, return a boolean array with the same batch dimensions as input.


    .. code-block:: python

        is_pos = is_positive(x())
        print(f'The operator is positive or not:\n{is_pos}')

    ::

        The operator is positive or not:
        False

    """
    mat = _type_transform(mat, "tensor")

    assert utils.check._is_square(mat), \
        f"The input matrix is not a square matrix: received shape {mat.shape}"

    return utils.check._is_positive(mat, eps).tolist()


def is_povm(
    set_op: Union[torch.Tensor, np.ndarray],
    eps: float = 1e-6
) -> Union[bool, List[bool]]:
    r"""Check if a set of operators forms a positive operator-valued measure (POVM).

    Args:
        set_op: A set of operators indexed by the first dimension
        eps: An optional tolerance value. Default value is 1e-6.

    Returns:
        whether the operators form a POVM.

    .. code-block:: python

        is_pov = is_povm(x())
        print(f'The operator is POVM or not:\n{is_pov}')

    ::

        The operator is POVM or not:
        False

    """
    set_op = _type_transform(set_op, "tensor")
    return utils.check._is_povm(set_op, eps)


def is_projector(
    mat: _ArrayLike, eps: float = 1e-6
) -> Union[bool, List[bool]]:
    r"""Verify whether ``mat`` is a projector.
        Support batch input.

    Args:
        mat: projector candidate :math:`P`
        eps: tolerance of error

    Returns:
        determine whether :math:`PP - P = 0`
        For batch input, return a boolean array with the same batch dimensions as input.

    .. code-block:: python

        is_proj = is_projector(x())
        print(f'The operator is projector or not:\n{is_proj}')

    ::

        The operator is projector or not:
        False

    """
    mat = _type_transform(mat, "tensor")

    assert utils.check._is_square(mat), \
        f"The input matrix is not a square matrix: received shape {mat.shape}"

    return utils.check._is_projector(mat, eps).tolist()


def is_ppt(density_op: _StateLike) -> Union[bool, List[bool]]:
    r"""Check if the input quantum state is PPT.
        Support batch input.

    Args:
        density_op: Density matrix form of the quantum state.

    Returns:
        Whether the input quantum state is PPT. 
        For batch input, return a boolean array with the same batch dimensions as input.

    .. code-block:: python

        is_pt = is_ppt(bell_state(2))
        print(f'The input quantum state is PPT or not:\n{is_pt}')

    ::

        The input quantum state is PPT or not:
        False

    """
    density_op = _type_transform(density_op, "state").density_matrix
    return utils.check._is_ppt(density_op).tolist()


def is_pvm(
    set_op: Union[torch.Tensor, np.ndarray],
    eps: float = 1e-6
) -> Union[bool, List[bool]]:
    r"""Check if a set of operators forms a projection-valued measure (PVM).

    Args:
        set_op: A set of operators indexed by the first dimension
        eps: An optional tolerance value. Default value is 1e-6.

    Returns:
        bool or List[bool]: True if the operators form a PVM; otherwise, False.

    .. code-block:: python

        is_pv = is_pvm(eye(4).reshape(1, 4, 4))
        print(f'The operator is PVM or not:\n{is_pv}')

    ::

        The operator is PVM or not:
        False

    """
    set_op = _type_transform(set_op, "tensor")
    return utils.check._is_pvm(set_op, eps)


def is_state_vector(
    vec: _ArrayLike, eps: float = 1e-6
) -> Union[bool, List[bool]]:
    r"""Verify whether ``vec`` is a legal quantum state vector.
        Support batch input.

    Args:
        vec: state vector candidate :math:`x`
        eps: tolerance of error, default to be `None` i.e. no testing for data correctness

    Returns:
        determine whether :math:`x^\dagger x = 1`

    .. code-block:: python

        st = torch.rand(5)
        is_st_vec = is_state_vector(st / st.norm())
        print(f'The vector is state vector or not:\n{is_st_vec}')

    ::

        The vector is state vector or not:
        True

    """
    vec = _type_transform(vec, "tensor")

    if vec.ndim == 1:
        vec = vec.view([-1, 1])
    else:
        vec = vec.view(list(vec.shape[:-2]) + [-1, 1])

    return utils.check._is_state_vector(vec, eps).tolist()


def is_unitary(
    mat: _ArrayLike, eps: Optional[float] = 1e-4
) -> Union[bool, List[bool]]:
    r"""Verify whether ``mat`` is a unitary.
        Support batch input.

    Args:
        mat: unitary candidate :math:`P`
        eps: tolerance of error

    Returns:
        determine whether :math:`PP^\dagger - I = 0`
        For batch input, return a boolean array with the same batch dimensions as input.

    .. code-block:: python

        is_unit = is_unitary(random_unitary(1))
        print(f'The operator is unitary or not:\n{is_unit}')

    ::

        The operator is unitary or not:
        True

    """
    mat = _type_transform(mat, "tensor")

    assert utils.check._is_square(mat), \
        f"The input matrix is not a square matrix: received shape {mat.shape}"

    return utils.check._is_unitary(mat, eps).tolist()



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The module that implements various backends of the state.
"""

r"""
States in QuAIRKit obeys the following broadcast rule:

| Batch size of Operators | Batch size of Input State | Batch size of Output State |
|:-----------------------:|:-------------------------:|:--------------------------:|
|           1             |             1             |             1            |
|           n             |             1             |             n            |
|           1             |             n             |             n            |
|           n             |             n             |             n            |
|           m             |             n             |          (m, n)          |

Note that the last rule currently is only applicable to the `evolve_keep_dim` method.
"""

import warnings
from abc import ABC, abstractmethod
from typing import List, Optional, Tuple, Union

import numpy as np
import torch

from ... import Hamiltonian, base, utils


class State(ABC):
    r"""The abstract base class for quantum states in QuAIRKit.

    Args:
        data: tensor array for quantum state(s).
        sys_dim: a list of dimensions for each system.
        system_seq: the system order of this state. Defaults to be from 0 to n - 1.

    """
    def __init__(self, data: torch.Tensor, sys_dim: List[int], 
                 system_seq: Optional[List[int]] = None):
        self._data, self._sys_dim = data.contiguous().to(dtype=base.get_dtype(), device=base.get_device()), sys_dim
        self._system_seq = list(range(len(sys_dim))) if system_seq is None else system_seq

        self.is_swap_back = True # TODO: depreciated, will be removed in the future
        self._keep_dim = False
    
    @abstractmethod
    def __getitem__(self, key: Union[int, slice]) -> 'State':
        r"""Indexing of the State class
        """
    
    @abstractmethod
    def prob_select(self, outcome_idx: torch.Tensor, prob_idx: int = -1) -> 'State':
        r"""Indexing probability outcome from the State batch along the given dimension.
        
        Args:
            outcome_idx: the 1-D tensor containing the outcomes to index
            prob_idx: the int that indexing which probability distribution. Defaults to be the last distribution.

        Returns:
            States that are selected by the specific probability outcomes.
        
        """
        
    @abstractmethod
    def expand(self, batch_dim: List[int]) -> 'State':
        r"""Expand the batch dimension of the State.
        
        Args:
            batch_dim: the new batch dimension
            
        Note:
            See torch.expand() for more information about expand. 
            This expand function, however, may change the format of _data to keep save memory.
        
        """

    def expand_as(self, other: Union['State', torch.Tensor]) -> 'State':
        r"""Expand this tensor to the same size as other.
        
        Args:
            other: the state/tensor to be expanded to. Note that if other is a
             torch.Tensor, the batch dimension of the state will be expanded to the same size as other.
        
        Note:
            See torch.Tensor.expand_as() for more information about expand.
            This expand function, however, may change the format of _data to keep save memory.
         
        """
        other_batch_dim = other.batch_dim if isinstance(other, State) else list(other.size())
        return self.expand(other_batch_dim)

    def __copy__(self) -> 'State':
        return self.clone()

    def __str__(self) -> str:
        split_line = "\n-----------------------------------------------------\n"
        s = f"{split_line} Backend: {self.backend}\n"
        s += f" System dimension: {self._sys_dim}\n"
        s += f" System sequence: {self._system_seq}\n"

        data = np.round(self.numpy(), decimals=2)
        interrupt_num = 5
        if not self.batch_dim:
            s += str(data.squeeze(0))
            s += split_line
            return s

        s += f" Batch size: {self.batch_dim}\n"
        for i, mat in enumerate(data):
            s += f"\n # {i}:\n{mat}"
            
            if i > interrupt_num:
                break_line = ("\n----------skipped for the rest of " + 
                              f"{list(data.shape)[0] - interrupt_num} states----------\n")
                s+= break_line
                return s
        
        s += split_line
        return s

    def __matmul__(self, other: 'State') -> torch.Tensor:
        r"""Matrix product between the density matrix representation of two states

        Args:
            other: a quantum state

        Returns:
            the product of these two states

        """
        return torch.matmul(self.density_matrix, other.density_matrix)

    @property
    @abstractmethod
    def backend() -> str:
        r"""The backend of this state.
        """

    @staticmethod
    @abstractmethod
    def check(data: torch.Tensor, sys_dim: Union[int, List[int]], eps: Optional[float] = 1e-4) -> int:
        r"""Assert whether the input data is valid for the specific State class.

        Args:
            data: tensor array for quantum state(s).
            sys_dim: (list of) dimension(s) of the systems, can be a list of integers or an integer.
            eps: the tolerance for the numerical check. Can be None means the check is overridden.

        Returns:
            the number of systems.

        """

    @property
    def data(self) -> torch.Tensor:
        r"""The data tensor of this state
        """
        warnings.warn(
            'The data property is depreciated, use ket or density_matrix instead', DeprecationWarning)
        return self._data
    
    def _joint_probability(self, prob_idx: List[int]) -> torch.Tensor:
        r"""The joint probability distribution of these states' occurrences
        
        Args:
            prob_idx: the indices of probability distributions

        """
        result_prob = torch.tensor(1.0)
        for idx in sorted(prob_idx):
            selected_prob = self._prob[idx]
            result_prob = result_prob.view(list(result_prob.shape) + [1] * (selected_prob.dim() - result_prob.dim()))
            result_prob = torch.mul(result_prob, selected_prob)
        return result_prob
    
    @property
    def probability(self) -> torch.Tensor:
        r"""The probability distribution(s) of these states' occurrences
        """
        return self._joint_probability(range(len(self._prob)))
    
    @property
    def _prob_dim(self) -> List[int]:
        r"""Current probability dimensions
        """
        return list(self._prob[-1].shape[-len(self._prob):]) if self._prob else []
    
    @property
    def batch_dim(self) -> List[int]:
        r"""The batch dimension of this state
        """
        return self._batch_dim + self._prob_dim
    
    @property
    def _squeeze_shape(self) -> List[int]:
        r"""The squeezed shape of this state batch
        """
        return [-1, int(np.prod(self._prob_dim))]
    
    @property
    def shape(self) -> torch.Size:
        r"""The shape of this state
        """
        return self._data.shape
    
    @property
    def dtype(self) -> torch.dtype:
        r"""The data type of this state
        """
        return self._data.dtype
    
    @property
    def device(self) -> torch.device:
        r"""The device of this state
        """
        return self._data.device

    def numel(self) -> int:
        r"""The number of elements in this data
        """
        return int(np.prod(self.batch_dim))

    @property
    def dim(self) -> int:
        r"""The dimension of this state
        """
        return self._data.shape[-1]
    
    @property
    def system_dim(self) -> List[int]:
        r"""The list of dimensions for each system
        """
        return self._sys_dim.copy()
    
    @system_dim.setter
    def system_dim(self, sys_dim: List[int]) -> None:
        r"""Set the system dimensions of the state
        
        Args:
            sys_dim: the target system dimensions.
        
        """
        self.reset_sequence()
        assert int(np.prod(sys_dim)) == self.dim, \
            f"The input system dim {sys_dim} does not match the original state dim {self.dim}."
        self._sys_dim = sys_dim.copy()
        self._system_seq = list(range(len(sys_dim)))
    
    def are_qubits(self) -> bool:
        r"""Whether all systems are qubits
        """
        return all(x == 2 for x in self._sys_dim)
    
    def are_qutrits(self) -> bool:
        r"""Whether all systems are qutrits
        """
        return all(x == 3 for x in self._sys_dim)
        
    @property
    def num_systems(self) -> int:
        r"""The number of systems
        """
        return len(self._sys_dim)

    @property
    def num_qubits(self) -> int:
        r"""The number of qubits of this state, when all systems are qubits
        """
        assert self.are_qubits, \
            f"Not all systems are qubits: received {self._sys_dim}"
        return len(self._sys_dim)

    @property
    @abstractmethod
    def ket(self) -> torch.Tensor:
        r"""The ket form of this state

        Note:
            If the state is pure, the ket form is the statevector representation.
            If the state is mixed, the ket form is the vectorization of its density matrix representation.

        """

    @property
    def bra(self) -> torch.Tensor:
        r"""Dagger of the ket form.
        """
        return self.ket.mH

    @property
    @abstractmethod
    def density_matrix(self) -> torch.Tensor:
        r"""The density matrix representation of this state.
        """
    
    @property
    def vec(self) -> torch.Tensor:
        r"""Vectorization of the state: :math:`\textrm{vec}(|i \rangle\langle j|)=|j, i \rangle`
        """
        data = self.density_matrix.view([1, -1, self.dim, self.dim])
        data = np.reshape(data, [data.shape[0], data.shape[1], -1, 1], order='F')
        return data.squeeze((0,1)).view(self.batch_dim + [-1, 1])
    
    @abstractmethod 
    def _trace(self, trace_idx: List[int]) -> 'State':
        r"""Partial trace of the state
        
        Args:
            trace_idx: the subsystem indices to be traced out.
        
        Returns:
            the partial trace of the state as a new state.
        
        """
        
    @abstractmethod
    def _transpose(self, transpose_idx: List[int]) -> 'State':
        r"""Partial transpose of the state
        
        Args:
            transpose_idx: the subsystem indices to be transposed.
        
        Returns:
            the transposed state as a new state.
        
        """

    @property
    def rank(self) -> List[int]:
        r"""The rank of the state.
        """
        dtype = self._data.dtype
        tol = 1e-8 if dtype == torch.complex64 else 1e-12
        tol *= self.dim
        return torch.linalg.matrix_rank(self.density_matrix, 
                                        tol=tol, hermitian=True).tolist()

    @abstractmethod
    def normalize(self) -> None:
        r"""Normalize this state
        """

    def numpy(self) -> np.ndarray:
        r"""Get the data in numpy.

        Returns:
            The numpy array of the data for the quantum state.
        """
        return self._data.detach().numpy()

    @abstractmethod
    def clone(self) -> 'State':
        r"""Return a copy of the quantum state.
        """
    
    @abstractmethod
    def fit(self, state_backend: str) -> torch.Tensor:
        r"""Convert the data to the specified backend.
        
        Args:
            state_backend: the target backend of the state. Available options see quairkit.Backend.

        Returns:
            a copy of the fitted data.
        
        """
    
    def to(self, dtype: str = None, device: str = None) -> 'State':
        r"""Change the property of the data tensor, and return a copy of this State

        Args:
            dtype: the new data type of the state.
            device: the new device of the state.

        """
        new_state = self.clone()
        new_state._data = new_state._data.to(dtype=dtype, device=device)
        return new_state
    
    @property
    def system_seq(self) -> List[int]:
        r"""The system order of this state
        """
        return self._system_seq.copy()

    @system_seq.setter
    @abstractmethod
    def system_seq(self, target_seq: List[int]) -> None:
        r"""Set the system order of the state.

        Args:
            target_seq: the target systems order.

        """
    
    def permute(self, target_seq: List[int]) -> 'State':
        r"""Permute the systems order of the state.
        
        Args:
            target_seq: the target systems order.
        
        """
        new_state = self.clone()
        new_state.system_seq = target_seq
        new_state._system_seq = list(range(self.num_systems))
        return new_state

    def reset_sequence(self) -> None:
        r"""reset the system order to default sequence i.e. from 0 to n - 1.
        """
        self.system_seq = list(range(self.num_systems))
    
    @abstractmethod
    def _evolve(self, unitary: torch.Tensor, sys_idx: List[int], on_batch: bool = True) -> None:
        r"""Evolve this state with unitary operators.

        Args:
            unitary: the unitary operator.
            sys_idx: the system indices to be acted on.
            on_batch: whether this unitary operator evolves on batch axis. Defaults to True.
        
        Note:
            The difference between `State.evolve` and `State._evolve` is that the former returns a new state 
            by calling an extra clone.
            
            Support the following transformation
            - (), () -> ()
            - (), (n) -> (n)
            - (n), () -> (n)
            - (n), (n) -> (n)
            - (m), (n) -> error
        
        """
        
    @abstractmethod
    def _evolve_keep_dim(self, unitary: torch.Tensor, sys_idx: List[int], on_batch: bool = True) -> None:
        r"""Evolve this state with unitary operators, while record the size of unitary batch
        
        Args:
            unitary: the unitary operator.
            sys_idx: the system indices to be acted on.
            on_batch: whether this unitary operator evolves on batch axis. Defaults to True.
            
        Note:
            Support the following transformation
            - (), () -> (1)
            - (m), () -> (m)
            - (), (n) -> (n, 1)
            - (m), (n) -> (n, m)
            - (n, m), (n) -> (n, m)
            - (l, m), (n) -> error
        """
    
    @abstractmethod
    def _evolve_ctrl(self, unitary: torch.Tensor, index: int, sys_idx: List[Union[int, List[int]]]) -> None:
        r"""Evolve this state with unitary operators controlled by a computational state
        
        Args:
            unitary: the unitary operator.
            index: the index of the computational state that activates the unitary.
            sys_idx: indices of the systems on which the whole unitary is applied. The first element in the list is 
                a list that gives the control system, while the remaining elements are int that give the applied system.
        
        Note:
            Support the following transformation
            - (), () -> ()
            - (), (n) -> (n)
            - (n), () -> (n)
            - (n), (n) -> (n)
            - (m), (n) -> error
        
        """
    
    @abstractmethod
    def _evolve_ctrl(self, unitary: torch.Tensor, index: int, sys_idx: List[Union[int, List[int]]]) -> None:
        r"""Evolve this state with unitary operators controlled by a computational state
        
        Args:
            unitary: the unitary operator.
            index: the index of the computational state that activates the unitary.
            sys_idx: indices of the systems on which the whole unitary is applied. The first element in the list is 
                a list that gives the control system, while the remaining elements are int that give the applied system.
        
        Note:
            Support the following transformation
            - (), () -> ()
            - (), (n) -> (n)
            - (n), () -> (n)
            - (n), (n) -> (n)
            - (m), (n) -> error
        
        """
    
    @abstractmethod
    def _transform(self, op: torch.Tensor, sys_idx: List[int], repr_type: str, on_batch: bool = True) -> None:
        r"""Apply a general linear operator to the state.

        Args:
            op: the input operator.
            sys_idx: the subsystem indices to be applied.
            repr_type: the representation type of input operator.
            on_batch: whether this operator evolves on batch axis. Defaults to True.

        Note:
            The difference between `State.transform` and `State._transform` is that the former returns a new state 
            by calling an extra clone.
            
            Support the following transformation
            - (), (n) -> (n)
            - (n), () -> (n)
            - (n), (n) -> (n)
            - (m), (n) -> error
        
        """
    
    @abstractmethod
    def _expec_val(self, obs: torch.Tensor, sys_idx: List[int]) -> torch.Tensor:
        r"""The expectation value of observables.
        
        Args:
            obs: the (list of) input observable.
            sys_idx: the system indices to be measured.
            
        Returns:
            the expectation value of the input observable for the quantum state.
        
        Note:
            Support the following transformation
            - (), (n) -> (n)
            - (n), () -> (n)
            - (n), (n) -> (n)
            - (m), (n) -> error
        
        """
    
    @abstractmethod
    def _expec_state(self, prob_idx: List[int]) -> 'State':
        r"""The expectation with respect to the specific probability distribution(s) of states
        
        Args:
            prob_idx: indices of probability distributions. Defaults to all distributions.
        """
    
    @abstractmethod
    def _measure(self, measure_op: torch.Tensor, sys_idx: List[int]) -> Tuple[torch.Tensor, 'State']:
        r"""Measure the quantum state with the measured operators.
        
        Args:
            measure_op: the measurement operators, where the first dimension is the number of operators.
        
        Returns:
            The probability and collapsed states of each measurement result.
        
        Note:
            Support the following transformation
            - (r), () -> (r)
            - (m, r), () -> (m, r)
            - (r), (n) -> (n, r)
            - (n, r), (n) -> (n, r)
            - (m, r), (n) -> error
        
        """
        
    def evolve(self, unitary: torch.Tensor, sys_idx: Optional[Union[int, List[int]]] = None) -> 'State':
        r"""Evolve this state with unitary operators.

        Args:
            unitary: the unitary operator.
            sys_idx: the system indices to be acted on. Defaults to all systems.

        Returns:
            the evolved state.

        Note:
            Evolve support the broadcast rule.
        
        """
        if sys_idx is None:
            sys_idx = list(range(self.num_systems))
        if isinstance(sys_idx, int):
            sys_idx = [sys_idx]
        
        applied_dim = unitary.shape[-1]
        assert applied_dim == int(np.prod((list_dim := [self._sys_dim[i] for i in sys_idx]))), \
            f"The unitary's dimension {applied_dim} does not match the dimensions of its acting systems {list_dim}"
        
        new_state = self.clone()
        new_state._evolve(unitary, sys_idx)
        return new_state

    def transform(self, op: torch.Tensor, sys_idx: Optional[Union[int, List[int]]] = None, 
                  repr_type: str = 'kraus') -> 'State':
        r"""Apply a general linear operator to the state.

        Args:
            op: the input operator.
            sys_idx: the qubit indices to be applied. Defaults to all systems.
            repr_type: the representation type of input operator. Defaults to 'kraus'.

        Returns:
            the transformed state.

        """
        # TODO add assertion for the dimension of input channels
        if sys_idx is None:
            sys_idx = list(range(self.num_systems))
        if isinstance(sys_idx, int):
            sys_idx = [sys_idx]
        
        # TODO add support for batched operators
        repr_type = repr_type.lower()
        if repr_type == 'kraus':
            if len(op.shape) == 2:
                op = op.unsqueeze(0)
            elif len(op.shape) > 3:
                raise NotImplementedError(
                    'consider the batched Kraus operators in the upcoming future')
        
        elif repr_type == 'choi':
            if len(op.shape) > 2:
                raise NotImplementedError(
                    'consider the batched Choi operators in the upcoming future')
        else:
            raise ValueError(
                f"Unsupported representation type {repr_type}: expected 'kraus' or 'choi'.")
        
        new_state = self.clone()
        new_state._transform(op, sys_idx, repr_type)
        return new_state
    
    def product_trace(self, trace_state: 'State', trace_idx: List[int]) -> 'State':
        r"""Partial trace over this state, when this state is a product state
        
        Args:
            trace_state: the state for the subsystem to be traced out.
            trace_idx: the subsystem indices to be traced out.
        
        Note:
            This function only works when the state is a product state represented by PureState
        
        """
        return self.trace(trace_idx)

    def expec_val(self, hamiltonian: Hamiltonian, shots: Optional[int] = 0, 
                  decompose: Optional[bool] = False) -> Union[torch.Tensor, List[torch.Tensor]]:
        r"""The expectation value of the observable with respect to the quantum state.

        Args:
            hamiltonian: Input observable.
            decompose: If decompose is ``True``, it will return the expectation value of each term.

        Returns:
            The expectation value (per term) of the input observable for the quantum state.
        
        Note:
            currently only run in qubit case.
        
        """
        if not self.are_qubits():
            raise NotImplementedError(
                f"Currently only support qubit computation in Hamiltonian tasks: received {self._sys_dim}")
        
        assert hamiltonian.n_qubits <= self.num_qubits, \
            f"The number of qubits in the Hamiltonian {hamiltonian.n_qubits} is greater the number of state qubits {self.num_qubits}"
        
        num_terms, list_qubits_idx = hamiltonian.n_terms, hamiltonian.sites
        list_coef, list_matrices = hamiltonian.coefficients, hamiltonian.pauli_words_matrix
        
        expec_val_each_term = []
        for i in range(num_terms):
            qubits_idx = list_qubits_idx[i]
            if qubits_idx == ['']:
                expec_val_each_term.append(list_coef[i] * self.trace())
                continue
            matrix = list_matrices[i]
            
            #TODO this assertion should be done by Hamiltonian, not State
            assert 2 ** len(qubits_idx) == matrix.shape[0], \
                f"The qubit index {qubits_idx} does not match the matrix dimension {matrix.shape[0]}"
            
            expec_val = self._expec_val(matrix.unsqueeze(0), qubits_idx).squeeze(-1).real * list_coef[i]
            expec_val_each_term.append(expec_val)
        
        expec_val_each_term = torch.stack(expec_val_each_term)
        return expec_val_each_term if decompose else torch.sum(expec_val_each_term, dim=0)
    
    def expec_state(self, prob_idx: Optional[Union[int, List[int]]] = None) -> 'State':
        r"""The expectation with respect to the specific probability distribution(s) of states
        
        Args:
            prob_idx: indices of probability distributions. Defaults to all distributions.
        
        Returns:
            The expected State obtained from the taken probability distributions.
        
        """
        if self._prob == []:
            return self.clone()
        num_prob = len(self._prob)
        
        if prob_idx is None:
            prob_idx = list(range(num_prob))
        elif isinstance(prob_idx, int):
            prob_idx = [prob_idx]
        else:
            prob_idx = [num_prob + idx if idx < 0 else idx for idx in sorted(prob_idx)]
        return self._expec_state(prob_idx)

    def measure(self, measured_op: Optional[torch.Tensor] = None, sys_idx: Optional[Union[int, List[int]]] = None, 
                is_povm: bool = False, keep_state: bool = False
                ) -> Union[torch.Tensor, Tuple[torch.Tensor, 'State']]:
        r"""Measure the quantum state

        Args:
            measured_op: measurement operator. Defaults to the computational basis.
            sys_idx: the system indices to be measured. Defaults to all systems.
            is_povm: whether the measurement is a POVM.
            keep_state: whether to keep the state after measurement.

        Returns:
            The probability of each measurement result. The corresponding collapsed state will also be returned if
            `is_povm` is False or `keep_state` is True.
        
        Raises:
            UserWarning: `is_povm` and `keep_state` cannot be both True,
            since a general POVM does not distinguish states.
        
        """
        # TODO add assertion for the dimension of measured operators
        if sys_idx is None:
            sys_idx = list(range(self.num_systems))
        if isinstance(sys_idx, int):
            sys_idx = [sys_idx]
        
        if measured_op is None:
            dim = int(np.prod([self._sys_dim[i] for i in sys_idx]))
            identity = torch.eye(dim, dtype=self.dtype, device=self.device).unsqueeze(-1)
            measured_op = identity @ identity.mH
        
        if is_povm:
            if keep_state:
                raise ValueError(
                    "`is_povm` and `keep_state` cannot be both True, " + 
                    "since a general POVM does not distinguish states.")
            return self._expec_val(measured_op, sys_idx)
            
        prob, collapsed_state = self._measure(measured_op, sys_idx)
        return (prob, collapsed_state) if keep_state else prob
    
    @abstractmethod
    def sqrt(self) -> torch.Tensor:
        r"""Matrix square root of the state.
        """
    
    @abstractmethod
    def log(self) -> torch.Tensor:
        r"""Matrix logarithm of the state.
        """
    
    def trace(self, trace_idx: Optional[Union[List[int], int]] = None) -> Union[torch.Tensor, 'State']:
        r"""(Partial) trace of the state
        
        Args:
            trace_idx: the subsystem indices to be traced out. Defaults to all systems.
        
        Returns:
            the trace of the state as a Tensor, or a new state if sys_idx is not None.
        """
        if trace_idx is None:
            return utils.linalg._trace(self.density_matrix, -2, -1)
        if isinstance(trace_idx, int):
            trace_idx = [trace_idx]
        assert max(trace_idx) < self.num_systems, \
            f"The trace index {trace_idx} should be smaller than number of systems {self.num_systems}"
            
        return self._trace(trace_idx)
    
    def transpose(self, transpose_idx: Optional[Union[List[int], int]] = None) -> 'State':
        r"""(Partial) transpose of the state
        
        Args:
            transpose_idx: the subsystem indices to be transposed. Defaults to all systems.
        
        Returns:
            the transpose of the state as a Tensor, or a new state if sys_idx is not None.
        """
        if transpose_idx is None:
            transpose_idx = list(range(self.num_systems))
        elif isinstance(transpose_idx, int):
            transpose_idx = [transpose_idx]
        assert max(transpose_idx) < self.num_systems, \
            f"The transpose index {transpose_idx} should be smaller than number of systems {self.num_systems}"
        
        return self._transpose(transpose_idx)



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The source file of the class for the distance.
"""

import torch

from ..core import Operator, State, utils

__all__ = ["TraceDistance", "StateFidelity"]

class TraceDistance(Operator):
    r"""The class of the loss function to compute the trace distance.

    This interface can make you using the trace distance as the loss function.

    Args:
        target_state: The target state to be used to compute the trace distance.
    
    .. code-block:: python

        from quairkit.database import one_state, bell_state, random_state
        
        # Single state example
        target_state = one_state(1)  # Define target
        input_state = bell_state(1)  # Define input
        trace_distance = TraceDistance(target_state)
        result = trace_distance(input_state)
        print('Single state distance:', result)  # Output: tensor(0.7071)
        
        # Batched states example
        target_state_batch = random_state(num_systems=1, size=2)  # Batch of 2 targets
        input_state_batch = random_state(num_systems=1, size=2)   # Batch of 2 inputs
        trace_distance = TraceDistance(target_state_batch)
        batch_result = trace_distance(input_state_batch)
        print('Batched distances:', batch_result)  # Output: tensor([0.7912, 0.7283])
    
    ::
    
        Single state distance: tensor(0.7071)
        Batched distances: tensor([0.7912, 0.7283])
    
    """
    def __init__(self, target_state: State):
        r"""Output target state.
        """
        super().__init__()
        self.target_state = target_state
        
    def __call__(self, state: State) -> torch.Tensor:
        r"""Same as forward of Neural Network
        """
        return self.forward(state)

    def forward(self, state: State) -> torch.Tensor:
        r"""Compute the trace distance between the input state and the target state.

        The value computed by this function can be used as a loss function to optimize.

        Args:
            state: The input state which will be used to compute the trace distance with the target state.

        Raises:
            NotImplementedError: The backend is wrong or not supported.

        Returns:
            The trace distance between the input state and the target state.
        
        """
        assert (
            state.num_systems == self.target_state.num_systems
        ), f"The number of systems does not match: received {state.num_systems}, expect {self.target_state.num_systems}."

        if not (state.backend == self.target_state.backend == 'state_vector'):
            return utils.qinfo._trace_distance(
                state.density_matrix, self.target_state.density_matrix
            )

        inner_prod = torch.linalg.vecdot(state.ket.squeeze(-1), self.target_state.ket.squeeze(-1))
        fidelity = torch.abs(inner_prod) ** 2
        return torch.sqrt(1 - fidelity)

class StateFidelity(Operator):
    r"""The class of the loss function to compute the state fidelity.

    This interface can make you using the state fidelity as the loss function.

    Args:
        target_state: The target state to be used to compute the state fidelity.
        
    .. code-block:: python
    
            from quairkit.database import one_state, bell_state, random_state
        
            # Single state example
            target_state = one_state(1)  # Define target
            input_state = bell_state(1)  # Define input
            fidelity_calculator = StateFidelity(target_state)
            result = fidelity_calculator(input_state)
            print('Single state fidelity:', result)  # Output: tensor(0.7071)
            
            # Batched states example
            target_batch = random_state(num_systems=1, size=2)  # Batch of 2 targets
            input_batch = random_state(num_systems=1, size=2)   # Batch of 2 inputs
            fidelity_calculator = StateFidelity(target_batch)
            batch_result = fidelity_calculator(input_batch)
            print('Batched fidelities:', batch_result)  # Output: tensor([0.5658, 0.7090])
                    
        ::
        
            Single state fidelity: tensor(0.7071)
            Batched fidelities: tensor([0.5658, 0.7090])

    """
    def __init__(self, target_state: State):
        r"""Output target state.
        """
        super().__init__()
        self.target_state = target_state
        
    def __call__(self, state: State) -> torch.Tensor:
        r"""Same as forward of Neural Network
        """
        return self.forward(state)

    def forward(self, state: State) -> torch.Tensor:
        r"""Compute the state fidelity between the input state and the target state.

        The value computed by this function can be used as a loss function to optimize.

        Args:
            state: The input state which will be used to compute the state fidelity with the target state.

        Raises:
            NotImplementedError: The backend is wrong or not supported.

        Returns:
            The state fidelity between the input state and the target state.
        
        """
        assert (
            state.num_systems == self.target_state.num_systems
        ), f"The number of systems does not match: received {state.num_systems}, expect {self.target_state.num_systems}."
        
        if not (state.backend == self.target_state.backend == 'state_vector'):
            return utils.qinfo._state_fidelity(
                state.density_matrix, self.target_state.density_matrix
            )
        inner_prod = torch.linalg.vecdot(state.ket.squeeze(-1), self.target_state.ket.squeeze(-1))
        return torch.abs(inner_prod)



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
QuAIRKit
========

QuAIRKit is a Python research framework for quantum computing, quantum information,
and quantum machine learning algorithm development. It focuses on flexible design,
real-time simulation and rapid verification of quantum and classical algorithms.

Installation
------------

The minimum Python environment for QuAIRKit is ``3.8``.
We recommend installing QuAIRKit with Python ``3.10``.

.. code-block:: bash

    conda create -n quair python=3.10
    conda activate quair
    conda install jupyter notebook

We recommend the following way of installing QuAIRKit with pip,

.. code-block:: bash

    pip install quairkit

or download all the files and finish the installation locally,

.. code-block:: bash

    git clone https://github.com/QuAIR/QuAIRKit
    cd QuAIRKit
    pip install -e .

Functionality
-------------

- Quantum neural network algorithm simulation
- Quantum circuit simulation & visualization
- Quantum channel simulation
- Quantum algorithm/information tools

Modules
-------

``quairkit``: QuAIRKit source code

- ``ansatz``: module of circuit templates
- ``database``: module of useful matrices & sets
- ``operator``: module of quantum operators
- ``qinfo``: library of quantum algorithms & information tools
- ``circuit``: quantum circuit interface

Tutorials
---------

Check out the tutorial folder on `GitHub <https://github.com/QuAIR/QuAIRKit>`_ for more information.

Relations with Paddle Quantum
-----------------------------

`Paddle Quantum <https://github.com/PaddlePaddle/Quantum>`_ is the world's first cloud-integrated
quantum machine learning platform based on Baidu PaddlePaddle. As most contributors to this project
are also contributors to Paddle Quantum, QuAIRKit incorporates key architectural elements and
interface designs from its predecessor. QuAIRKit focuses more on providing specialized tools and
resources for researchers and developers engaged in cutting-edge quantum algorithm design and
theoretical explorations in quantum information science.
"""

import os
os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"

from .core import set_backend, set_device, set_dtype, set_seed
from .core import get_backend, get_device, get_dtype, get_seed, get_float_dtype
from .core import Hamiltonian, State, Operator, to_state
from . import ansatz
from . import database
from . import operator
from . import loss
from . import qinfo
from .circuit import Circuit
from . import application

name = "quairkit"
__version__ = "0.4.0"


def print_info() -> None:
    r"""Print the information of QuAIRKit, its dependencies and current environment.

    """
    import torch
    import numpy
    import scipy
    import matplotlib
    print("\n---------VERSION---------")
    print("quairkit:", __version__)
    print("torch:", torch.__version__)
    if torch.cuda.is_available():
        print("torch cuda:", torch.version.cuda)
    print("numpy:", numpy.__version__)
    print("scipy:", scipy.__version__)
    print("matplotlib:", matplotlib.__version__)

    import platform
    print("---------SYSTEM---------")
    print("Python version:", platform.python_version())
    print("OS:", platform.system())
    print("OS version:", platform.version())

    
    import subprocess
    import re
    # stack overflow #4842448
    print("---------DEVICE---------")
    if platform.system() == "Windows":
        cpu_name = platform.processor()
        
    elif platform.system() == "Darwin":
        model = subprocess.check_output(["/usr/sbin/sysctl", "-n", "machdep.cpu.brand_string"]).strip()
        cpu_name = model.decode('utf-8')

    elif platform.system() == "Linux":
        command = "cat /proc/cpuinfo"
        all_info = subprocess.check_output(command, shell=True).decode().strip()
        for line in all_info.split("\n"):
            if "model name" in line:
                cpu_name = re.sub(".*model name.*:", "", line, 1)
    print("CPU:", cpu_name)

    if torch.cuda.is_available():
        print("GPU: (0)", torch.cuda.get_device_name(0))
        for i in range(torch.cuda.device_count() - 1):
            print(f"     ({i})", torch.cuda.get_device_name(i))



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The module of the quantum channels.
"""

from .base import Channel
from .common import (AmplitudeDamping, BitFlip, BitPhaseFlip, Depolarizing,
                     GeneralizedAmplitudeDamping, GeneralizedDepolarizing,
                     PauliChannel, PhaseDamping, PhaseFlip, ReplacementChannel,
                     ResetChannel, ThermalRelaxation)
from .custom import ChoiRepr, KrausRepr, StinespringRepr



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The library of functions in quantum information.
"""


import math
from typing import Callable, List, Optional, Tuple, Union

import numpy as np
import torch

from .. import utils


def _choi_to_kraus(choi_repr: torch.Tensor, tol: float) -> List[torch.Tensor]:
    r"""Transform the Choi representation to the Kraus representation
    """
    ndim = int(math.sqrt(choi_repr.shape[0]))
    w, v = torch.linalg.eigh(choi_repr)

    # add abs to make eigvals safe
    w = torch.abs(w)
    l_cut = 0
    for l in range(len(w) - 1, -1, -1):
        if torch.sum(torch.abs(w[l:])) / torch.sum(torch.abs(w)) > 1 - tol:
            l_cut = l
            break
    return torch.stack([(v * torch.sqrt(w))[:, l].reshape([ndim, ndim]).T for l in range(l_cut, ndim**2)])


def _kraus_to_stinespring(kraus_repr: List[torch.Tensor]) -> torch.Tensor:
    r"""Transform the Kraus representation to the Stinespring representation
    """
    # TODO simplify logic
    j_dim = kraus_repr.shape[0]
    i_dim = kraus_repr.shape[1]
    stinespring_repr = kraus_repr.permute([1, 0, 2])
    return stinespring_repr.reshape([i_dim * j_dim, i_dim])


def _choi_to_stinespring(choi_repr: torch.Tensor, tol: float) -> List[torch.Tensor]:
    r"""Transform the Choi representation to the Stinespring representation
    """
    # TODO: need a more straightforward transformation
    return _kraus_to_stinespring(_choi_to_kraus(choi_repr, tol))


def _kraus_to_choi(kraus_repr: List[torch.Tensor]) -> torch.Tensor:
    r"""Transform the Kraus representation to the Choi representation
    """
    ndim = kraus_repr[0].shape[0]
    y = kraus_repr[0]
    kraus_oper_tensor = torch.reshape(
        torch.cat([torch.kron(x, x.conj().T.contiguous()) for x in kraus_repr]),
        shape=[len(kraus_repr), ndim, -1]
    )
    choi_repr = torch.sum(kraus_oper_tensor, axis=0).reshape([ndim for _ in range(4)]).permute([2, 1, 0, 3])
    return choi_repr.permute([0, 2, 1, 3]).reshape([ndim * ndim, ndim * ndim])


def _stinespring_to_kraus(stinespring_repr: torch.Tensor) -> torch.Tensor:
    r"""Transform the Stinespring representation to the Kraus representation
    """
    i_dim = stinespring_repr.shape[1]
    j_dim = stinespring_repr.shape[0] // i_dim
    return stinespring_repr.reshape([i_dim, j_dim, i_dim]).permute([1, 0, 2])


def _stinespring_to_choi(stinespring_repr: torch.Tensor) -> torch.Tensor:
    r"""Transform the Stinespring representation to the Choi representation
    """
    # TODO: need a more straightforward transformation
    return _kraus_to_choi(_stinespring_to_kraus(stinespring_repr))


def _trace_distance(rho: torch.Tensor, sigma: torch.Tensor) -> torch.Tensor:
    list_eigval = torch.linalg.eigvalsh(rho - sigma)
    dist = 0.5 * torch.sum(torch.abs(list_eigval), axis=-1)
    return dist.view(rho.shape[:-2] or sigma.shape[:-2])


def _state_fidelity(rho: torch.Tensor, sigma: torch.Tensor) -> torch.Tensor:
    sqrt_rho = utils.linalg._sqrtm(rho)
    fidelity = utils.linalg._trace(utils.linalg._sqrtm(sqrt_rho @ sigma @ sqrt_rho)).real
    return fidelity.view(rho.shape[:-2] or sigma.shape[:-2])


def _gate_fidelity(U: torch.Tensor, V: torch.Tensor) -> torch.Tensor:
    fidelity = torch.abs(utils.linalg._trace(U @ V.mH)) / U.shape[-1]
    return fidelity.view(U.shape[:-2] or V.shape[:-2])

def _purity(rho: torch.Tensor) -> torch.Tensor:
    product = torch.matmul(rho, rho)
    diagonal_elements = product.diagonal(dim1=-2, dim2=-1)
    purity = diagonal_elements.sum(dim=-1, keepdim=True).real
    return purity.view(rho.shape[:-2])


def _von_neumann_entropy(rho: torch.Tensor, base: Optional[Union[int, float]] = 2) -> torch.Tensor:
    entropy = -1 * math.log(math.e, base) * utils.linalg._trace(rho @ utils.linalg._logm(rho)).real
    return entropy.view(rho.shape[:-2])


def _relative_entropy(rho: torch.Tensor, sig: torch.Tensor, base: Optional[Union[int, float]] = 2) -> torch.Tensor:
    entropy = math.log(math.e, base) * utils.linalg._trace(rho @ utils.linalg._logm(rho) - rho @ utils.linalg._logm(sig)).real
    return entropy.view(rho.shape[:-2] or sig.shape[:-2])


def _negativity(density_op: torch.Tensor) -> torch.Tensor:
    half_dim = math.isqrt(density_op.shape[-1])
    density_op_transposed = utils.linalg._transpose_1(density_op, half_dim)
    
    # Calculate eigenvalues for the entire batch
    eigen_vals = torch.linalg.eigvalsh(density_op_transposed)
    
    # Calculate negativity for each density operator in the batch
    negativities = torch.sum(torch.where(eigen_vals <= -1e-16 * density_op.shape[-1], eigen_vals, 0.), dim=-1)
    return torch.abs(negativities).unsqueeze(-1).view(density_op.shape[:-2])


def _logarithmic_negativity(density_op: torch.Tensor) -> torch.Tensor:
    return torch.log2(2 * _negativity(density_op) + 1)


# TODO: double value & insufficient precision in matlab
def _diamond_norm(choi_matrix: torch.Tensor, dim_io: Optional[Union[int, Tuple[int, int]]] = None, **kwargs) -> float:
    import cvxpy

    if dim_io is None:    # Default to dim_in == dim_out
        dim_in = dim_out = int(math.sqrt(choi_matrix.shape[0]))
    elif isinstance(dim_io, tuple):
        dim_in = int(dim_io[0])
        dim_out = int(dim_io[1])
    elif isinstance(dim_io, int):
        dim_in = dim_io
        dim_out = dim_io
    else:
        raise TypeError('"dim_io" should be "int" or "tuple".')
    kron_size = dim_in * dim_out

    # Cost function : Trace( \Omega @ Choi_matrix )
    rho = cvxpy.Variable(shape=(dim_in, dim_in), complex=True)
    omega = cvxpy.Variable(shape=(kron_size, kron_size), complex=True)
    identity = np.eye(dim_out)

    # \rho \otimes 1 \geq \Omega
    cons_matrix = cvxpy.kron(rho, identity) - omega
    cons = [
        rho >> 0,
        rho.H == rho,
        cvxpy.trace(rho) == 1,

        omega >> 0,
        omega.H == omega,

        cons_matrix >> 0
    ]

    obj = cvxpy.Maximize(2 * cvxpy.real((cvxpy.trace(omega @ choi_matrix))))
    prob = cvxpy.Problem(obj, cons)

    return prob.solve(**kwargs)


def _create_choi_repr(
    linear_map: Callable[[torch.Tensor], torch.Tensor],
    input_dim: int,
    input_dtype: torch.dtype,
) -> torch.Tensor:
    Choi_matrix = 0

    for j in range(input_dim):
        for i in range(input_dim):
            E_ij = torch.zeros(
                input_dim,
                input_dim,
                dtype=input_dtype,
            )
            E_ij[i, j] = 1
            Choi_matrix += E_ij.kron(linear_map(E_ij))
    return Choi_matrix


def _decomp_1qubit(
    unitary: torch.Tensor, return_global: bool = False
) -> Tuple[torch.Tensor, ...]:
    a00 = unitary[..., 0, 0]
    a10 = unitary[..., 1, 0]
    a11 = unitary[..., 1, 1]

    alpha = (torch.angle(a00) + torch.angle(a11)) / 2
    beta = torch.angle(a10) - torch.angle(a00)
    delta = torch.angle(a11) - torch.angle(a10)
    gamma = 2 * torch.acos(a00 / torch.exp(1j * (alpha - beta / 2 - delta / 2))).real

    if return_global:
        return alpha, beta, gamma, delta
    else:
        return beta, gamma, delta


def _decomp_ctrl_1qubit(
    unitary: torch.Tensor,
) -> Tuple[torch.Tensor, ...]:
    alpha, beta, gamma, delta = _decomp_1qubit(unitary, return_global=True)

    ry = lambda theta: torch.tensor(
        [
            [torch.cos(theta / 2), -torch.sin(theta / 2)],
            [torch.sin(theta / 2), torch.cos(theta / 2)],
        ],
        dtype=unitary.dtype,
    )

    rz = lambda theta: torch.tensor(
        [[torch.exp(-1j * theta / 2), 0], [0, torch.exp(1j * theta / 2)]]
    )
    A = rz(beta) @ ry(gamma / 2)
    B = ry(-gamma / 2) @ rz(-(delta + beta) / 2)
    C = rz((delta - beta) / 2)

    return alpha, A, B, C


def _stab_renyi(
    density: torch.Tensor,
    alpha: torch.Tensor,
    num: int,
    indices: torch.Tensor,
    pauli: torch.Tensor,
    n: int,
) -> torch.Tensor:

    density = density.to(torch.complex128)

    # Add new dimensions for broadcasting
    # density.unsqueeze(1)   Shape: (m, 1, 2^n, 2^n)
    # pauli.unsqueeze(0)   Shape: (1, 4^n, 2^n, 2^n)

    Chi_func = (
        utils.linalg._trace(
            density.unsqueeze(1) @ pauli.unsqueeze(0), axis1=-2, axis2=-1
        ).real
    ) ** 2 / (
        2**n
    )  # shape (m,4^n)

    def _compute_p_norm_vector(vec: torch.Tensor, p: torch.Tensor) -> torch.Tensor:
        if vec.dim() == 1:
            vec = vec.unsqueeze(0)  # Convert to (1, P)
        # Compute the 1/2-norm
        p_norm = vec.abs().pow(p).sum(dim=1).pow(1 / p)
        return p_norm

    alpha_norm = _compute_p_norm_vector(Chi_func, alpha)

    # Initialize a result tensor with NaN values
    result_counts = torch.full((num,), float("nan"), dtype=torch.float)

    # Fill in the counts for density matrices
    result_counts[indices] = alpha_norm.to(result_counts.dtype)

    # Compute the alpha-stabilizer renyi entropy
    return alpha / (1 - alpha) * torch.log2(result_counts) - math.log2(2**n)


def _stab_nullity(
    unitary: torch.Tensor,
    num_unitary: int,
    unitary_indices: torch.Tensor,
    pauli: torch.Tensor,
    n: int,
) -> torch.Tensor:

    # Expand dimensions for broadcasting
    unitary_expanded = unitary.unsqueeze(1).unsqueeze(
        2
    )  # Shape (batchsize, 1, 1, 2^n, 2^n)
    unitary_d_expanded = (
        utils.linalg._dagger(unitary).unsqueeze(1).unsqueeze(2)
    )  # Shape (batchsize, 1, 1, 2^n, 2^n)
    pauli_expanded_i = pauli.unsqueeze(0).unsqueeze(2)  # Shape (1, 4^n, 1, 2^n, 2^n)
    pauli_expanded_k = pauli.unsqueeze(0).unsqueeze(1)  # Shape (1, 1, 4^n, 2^n, 2^n)

    # Third compute the Pauli function of unitary varing Pauli
    paulifunc = utils.linalg._trace(
        pauli_expanded_i @ unitary_expanded @ pauli_expanded_k @ unitary_d_expanded,
        axis1=-2,
        axis2=-1,
    ).real / (
        2**n
    )  # Shape (m, 4^n, 4^n, 2^n, 2^n) to Shape (m, 4^n, 4^n)

    # Count the number of +1 and -1 of paulifunc for every unitary to get the s(U)
    # Define a small tolerance
    tolerance = 1e-6
    # Create the mask with tolerance
    mask = (torch.abs(paulifunc - 1) < tolerance) | (
        torch.abs(paulifunc + 1) < tolerance
    )

    # Count the occurrences of True values in the mask
    counts = torch.sum(mask, dim=(-2, -1), dtype=torch.float)

    # Initialize a result tensor with NaN values
    result_counts = torch.full((num_unitary,), float("nan"), dtype=torch.float)

    # Fill in the counts for unitary matrices
    result_counts[unitary_indices] = counts

    # Compute the unitary-stabilizer nullity
    return 2 * n - torch.log2(result_counts)


def _mana_state(state: torch.Tensor, A: torch.Tensor, dim: int) -> torch.Tensor:
    state = state.to(torch.complex128)
    # Call trace function to get Wigner function
    # If state has shape (d, d), expand its dimensions to (1, d, d)
    state = state.unsqueeze(0) if state.dim() == 2 else state
    W = 1 / dim * utils.linalg._trace(state.unsqueeze(1) @ A.unsqueeze(0)).real

    return torch.log2((torch.abs(W).sum(dim=-1)))


def _mana_channel(
    channel: torch.Tensor,
    A_a: torch.Tensor,
    A_b: torch.Tensor,
    out_dim: int,
    in_dim: int,
) -> torch.Tensor:
    channel = channel.to(torch.complex128)
    # Compute the Kronecker product
    A_kron = torch.einsum("aij,bkl->abikjl", A_a.transpose(1, 2), A_b).reshape(
        in_dim**2, out_dim**2, in_dim * out_dim, in_dim * out_dim
    )
    # compute the wigner function of a quantum channel
    channel = channel.unsqueeze(0) if channel.dim() == 2 else channel
    W = (
        1
        / out_dim
        * utils.linalg._trace(
            channel.unsqueeze(1).unsqueeze(2) @ A_kron.unsqueeze(0)
        ).real
    )
    # Compute the mana of channels
    return torch.log2(torch.max(torch.sum(torch.abs(W), dim=-1), dim=-1).values)


def _stab_renyi(
    density: torch.Tensor,
    alpha: torch.Tensor,
    num: int,
    indices: torch.Tensor,
    pauli: torch.Tensor,
    n: int,
) -> torch.Tensor:

    density = density.to(torch.complex128)

    # Add new dimensions for broadcasting
    # density.unsqueeze(1)   Shape: (m, 1, 2^n, 2^n)
    # pauli.unsqueeze(0)   Shape: (1, 4^n, 2^n, 2^n)

    Chi_func = (
        utils.linalg._trace(
            density.unsqueeze(1) @ pauli.unsqueeze(0), axis1=-2, axis2=-1
        ).real
    ) ** 2 / (
        2**n
    )  # shape (m,4^n)

    def _compute_p_norm_vector(vec: torch.Tensor, p: torch.Tensor) -> torch.Tensor:
        if vec.dim() == 1:
            vec = vec.unsqueeze(0)  # Convert to (1, P)
        # Compute the 1/2-norm
        p_norm = vec.abs().pow(p).sum(dim=1).pow(1 / p)
        return p_norm

    alpha_norm = _compute_p_norm_vector(Chi_func, alpha)

    # Initialize a result tensor with NaN values
    result_counts = torch.full((num,), float("nan"), dtype=torch.float)

    # Fill in the counts for density matrices
    result_counts[indices] = alpha_norm.to(result_counts.dtype)

    # Compute the alpha-stabilizer renyi entropy
    return alpha / (1 - alpha) * torch.log2(result_counts) - math.log2(2**n)


def _stab_nullity(
    unitary: torch.Tensor,
    num_unitary: int,
    unitary_indices: torch.Tensor,
    pauli: torch.Tensor,
    n: int,
) -> torch.Tensor:

    # Expand dimensions for broadcasting
    unitary_expanded = unitary.unsqueeze(1).unsqueeze(
        2
    )  # Shape (batchsize, 1, 1, 2^n, 2^n)
    unitary_d_expanded = (
        utils.linalg._dagger(unitary).unsqueeze(1).unsqueeze(2)
    )  # Shape (batchsize, 1, 1, 2^n, 2^n)
    pauli_expanded_i = pauli.unsqueeze(0).unsqueeze(2)  # Shape (1, 4^n, 1, 2^n, 2^n)
    pauli_expanded_k = pauli.unsqueeze(0).unsqueeze(1)  # Shape (1, 1, 4^n, 2^n, 2^n)

    # Third compute the Pauli function of unitary varing Pauli
    paulifunc = utils.linalg._trace(
        pauli_expanded_i @ unitary_expanded @ pauli_expanded_k @ unitary_d_expanded,
        axis1=-2,
        axis2=-1,
    ).real / (
        2**n
    )  # Shape (m, 4^n, 4^n, 2^n, 2^n) to Shape (m, 4^n, 4^n)

    # Count the number of +1 and -1 of paulifunc for every unitary to get the s(U)
    # Define a small tolerance
    tolerance = 1e-6
    # Create the mask with tolerance
    mask = (torch.abs(paulifunc - 1) < tolerance) | (
        torch.abs(paulifunc + 1) < tolerance
    )

    # Count the occurrences of True values in the mask
    counts = torch.sum(mask, dim=(-2, -1), dtype=torch.float)

    # Initialize a result tensor with NaN values
    result_counts = torch.full((num_unitary,), float("nan"), dtype=torch.float)

    # Fill in the counts for unitary matrices
    result_counts[unitary_indices] = counts

    # Compute the unitary-stabilizer nullity
    return 2 * n - torch.log2(result_counts)


def _mana_state(state: torch.Tensor, A: torch.Tensor, dim: int) -> torch.Tensor:
    state = state.to(torch.complex128)
    # Call trace function to get Wigner function
    # If state has shape (d, d), expand its dimensions to (1, d, d)
    state = state.unsqueeze(0) if state.dim() == 2 else state
    W = 1 / dim * utils.linalg._trace(state.unsqueeze(1) @ A.unsqueeze(0)).real

    return torch.log2((torch.abs(W).sum(dim=-1)))


def _mana_channel(
    channel: torch.Tensor,
    A_a: torch.Tensor,
    A_b: torch.Tensor,
    out_dim: int,
    in_dim: int,
) -> torch.Tensor:
    channel = channel.to(torch.complex128)
    # Compute the Kronecker product
    A_kron = torch.einsum("aij,bkl->abikjl", A_a.transpose(1, 2), A_b).reshape(
        in_dim**2, out_dim**2, in_dim * out_dim, in_dim * out_dim
    )
    # compute the wigner function of a quantum channel
    channel = channel.unsqueeze(0) if channel.dim() == 2 else channel
    W = (
        1
        / out_dim
        * utils.linalg._trace(
            channel.unsqueeze(1).unsqueeze(2) @ A_kron.unsqueeze(0)
        ).real
    )
    # Compute the mana of channels
    return torch.log2(torch.max(torch.sum(torch.abs(W), dim=-1), dim=-1).values)

def _general_state_fidelity(rho: torch.Tensor, sigma: torch.Tensor) -> torch.Tensor:
    trace_norm_term = _state_fidelity(rho, sigma)
    general_fidelity = (trace_norm_term + torch.sqrt(
        (1 - utils.linalg._trace(rho)) * (1 - utils.linalg._trace(sigma))
    )).real
    return general_fidelity.view(rho.shape[:-2] or sigma.shape[:-2])

def _mutual_information(rho: torch.Tensor, dim_A: int, dim_B: int) -> torch.Tensor:
    system_dim = [dim_A, dim_B]
    rho_A = utils.linalg._partial_trace(rho, [1], system_dim)
    rho_B = utils.linalg._partial_trace(rho, [0], system_dim)
    return _von_neumann_entropy(rho_A) + _von_neumann_entropy(rho_B) - _von_neumann_entropy(rho)


def _link(
        JE: Tuple[torch.Tensor, str, List[int], List[int]],
        JF: Tuple[torch.Tensor, str, List[int], List[int]]
        ) -> Tuple[torch.Tensor, str, List[int], List[int]]:
    # FIXME: does not work for comb case, i.e., when PI -> OF and O -> I
    # Unpack variables
    JE_matrix, JE_entry_exit, JE_input_dims, JE_output_dims = JE
    JF_matrix, JF_entry_exit, JF_input_dims, JF_output_dims = JF

    JE_entry, JE_exit = JE_entry_exit.split('->')
    JF_entry, JF_exit = JF_entry_exit.split('->')

    # Calculate overlap
    overlap_subsystem = set(JE_exit).intersection(set(JF_entry))

    # Generate index based on overlap for subsequent permutation
    new_index = list(range(len(JE_entry) + len(JE_exit)))
    overlap_indices = [JE_exit.index(x) + len(JE_entry) for x in overlap_subsystem]
    exchange_indices = list(range(len(overlap_subsystem)))

    for old, new in zip(overlap_indices, exchange_indices):
        new_index[new], new_index[old] = new_index[old], new_index[new]

    # Permute, partial transpose, and permute back
    JE_choi_permuted = utils.linalg._permute_systems(
        JE_matrix, 
        new_index, 
        dim_list=JE_input_dims + JE_output_dims
    )

    JE_transposed = utils.linalg._partial_transpose(
        JE_choi_permuted, [0],
        [2 ** len(overlap_subsystem), JE_choi_permuted.shape[0] // (2 ** len(overlap_subsystem))]
    )

    JE_choi_transposed = utils.linalg._permute_systems(
        JE_transposed, 
        new_index, 
        JE_input_dims + JE_output_dims
    )

    # Generate dictionaries for each system and its corresponding dimension
    JE_pairs = list(zip(JE_entry, JE_input_dims)) + list(zip(JE_exit, JE_output_dims))
    JF_pairs = list(zip(JF_entry, JF_input_dims)) + list(zip(JF_exit, JF_output_dims))

    JE_dim_dict = dict(JE_pairs)
    JF_dim_dict = dict(JF_pairs)

    # Remove overlap subsystem from dictionaries
    for letter in overlap_subsystem:
        JE_dim_dict.pop(letter, None)
        JF_dim_dict.pop(letter, None)

    # Get the non-overlapping dimensions
    non_overlap_dims_E = list(JE_dim_dict.values())
    non_overlap_dims_F = list(JF_dim_dict.values())

    # Multiply the two Choi matrices
    multiplication = (
        torch.kron(JE_choi_transposed.contiguous(), torch.eye(np.prod(non_overlap_dims_F))) 
        @ torch.kron(torch.eye(np.prod(non_overlap_dims_E)), JF_matrix.contiguous())
    )

    # Generate index for subsequent partial trace
    total_str = (
        JE_entry 
        + ''.join([c for c in JE_exit if c not in overlap_subsystem]) 
        + JF_entry 
        + JF_exit
    )
    
    index_dict = {char: idx for idx, char in enumerate(total_str)}
    for letter in overlap_subsystem:
        index_dict.pop(letter, None)

    # Calculate the partial trace over the joint system
    result_matrix = utils.linalg._partial_trace_discontiguous(
        multiplication, 
        list(index_dict.values())
    )

    # Permute the result matrix to the correct order
    combined_str = (
        JE_entry 
        + ''.join([c for c in JE_exit if c not in overlap_subsystem]) 
        + ''.join([c for c in JF_entry if c not in overlap_subsystem]) 
        + JF_exit
    )
    index_nonoverlap_dict = {char: idx for idx, char in enumerate(combined_str)}

    permute_str = (
        JE_entry 
        + ''.join([c for c in JF_entry if c not in overlap_subsystem]) 
        + ''.join([c for c in JE_exit if c not in overlap_subsystem]) 
        + JF_exit
    )
    permute_list = [index_nonoverlap_dict[char] for char in permute_str]

    all_dims = non_overlap_dims_E + non_overlap_dims_F
    permute_dims = [all_dims[i] for i in permute_list]

    result_matrix = utils.linalg._permute_systems(result_matrix, permute_list, permute_dims)

    # Generate the entry and exit string for the final Choi matrix
    entry_exit = (
        JE_entry 
        + ''.join([c for c in JF_entry if c not in overlap_subsystem]) 
        + '->' 
        + ''.join([c for c in JE_exit if c not in overlap_subsystem]) 
        + JF_exit
    )

    # Extract the input and output dimensions
    entry, exit = entry_exit.split('->')

    entry_index = [index_nonoverlap_dict[char] for char in entry]
    input_dims = [all_dims[i] for i in entry_index]

    exit_index = [index_nonoverlap_dict[char] for char in exit]
    output_dims = [all_dims[i] for i in exit_index]

    return result_matrix, entry_exit, input_dims, output_dims



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The library of common Hamiltonians.
"""

import torch

from ..core import Hamiltonian

__all__ = ["ising_hamiltonian", "xy_hamiltonian", "heisenberg_hamiltonian"]


def ising_hamiltonian(edges: torch.Tensor, vertices: torch.Tensor) -> Hamiltonian:
    r"""Compute the Ising Hamiltonian.

    .. math::

        \begin{align}
            H_{Ising}= \sum_{(u,v) \in E(u>v)}\gamma_{uv}Z_u Z_v + \sum_{k \in V}\beta_k X_k
        \end{align}

    Args:
        edges: A tensor E shape=[V, V], where E[u][v] is \gamma_{uv}.
        vertices: A tensor V shape=[V], where V[k] is \beta_{k}.

    Returns:
        Hamiltonian representation of the Ising Hamiltonian.

    Examples:
        .. code-block:: python

            import torch
            from quairkit.database.hamiltonian import ising_hamiltonian

            edges = torch.tensor([[0, 1, 0.5],
                                  [1, 0, 0.2],
                                  [0.5, 0.2, 0]])
            vertices = torch.tensor([0.3, 0.4, 0.1])
            hamiltonian = ising_hamiltonian(edges, vertices)
            print(f'The Ising_Hamiltonian is \n {hamiltonian}')

        ::

            The Ising_Hamiltonian is 
            1.0 Z0, Z1
            0.5 Z0, Z2
            0.20000000298023224 Z1, Z2
            0.30000001192092896 X0
            0.4000000059604645 X1
            0.10000000149011612 X2
    """
    h_list = []
    shape_of_edges = edges.shape

    assert len(shape_of_edges) == 2, \
        f'The input variable edges should be a 2-dimension torch.tensor, but receive {len(shape_of_edges)}-dimension'

    edges_param_list = edges.tolist()
    vertices_param_list = vertices.tolist()

    for i in range(shape_of_edges[0]):
        for j in range(i + 1, shape_of_edges[1]):
            para = edges_param_list[i][j]
            if para == 0:
                continue
            h_list.append((para, f'Z{i},Z{j}'))

    for i in range(vertices.shape[0]):
        para = vertices_param_list[i]
        if para == 0:
            continue
        h_list.append((para, f'X{i}'))

    return Hamiltonian(h_list)


def xy_hamiltonian(edges: torch.Tensor) -> Hamiltonian:
    r"""Compute the XY Hamiltonian.

    .. math::

        \begin{align}
            H_{XY}= \sum_{(u,v) \in E(u>v)}(\alpha_{uv}X_u X_v + \beta_{uv}Y_u Y_v)
        \end{align}

    Args:
        edges: A tensor E shape=[2, V, V], where E[0][u][v] is \alpha_{uv} and E[1][u][v] is \beta_{uv}.

    Returns:
        Hamiltonian representation of the XY Hamiltonian.

    Examples:
        .. code-block:: python

            import torch
            from quairkit.database.hamiltonian import xy_hamiltonian

            edges = torch.tensor([
                [
                    [0, 0.7, 0],
                    [0.7, 0, 0.2],
                    [0, 0.2, 0]
                ],
                [
                    [0, 0.5, 0],
                    [0.5, 0, 0.3],
                    [0, 0.3, 0]
                ]
            ])
            H_XY = xy_hamiltonian(edges)
            print(f'The XY Hamiltonian is:\n{H_XY}')

        ::

            The XY Hamiltonian is:
            0.699999988079071 X0, X1
            0.5 Y0, Y1
            0.20000000298023224 X1, X2
            0.30000001192092896 Y1, Y2
    """
    h_list = []
    shape_of_edges = edges.shape

    assert len(shape_of_edges) == 3, \
        f'The input variable edges should be a 3-dimension torch.tensor, but receive {len(shape_of_edges)}-dimension'

    edges_param_list = edges.tolist()

    for i in range(shape_of_edges[1]):
        for j in range(i + 1, shape_of_edges[1]):
            alpha = edges_param_list[0][i][j]
            beta = edges_param_list[1][i][j]
            if alpha == 0 and beta == 0:
                continue
            h_list.extend(((alpha, f'X{i},X{j}'), (beta, f'Y{i},Y{j}')))
    return Hamiltonian(h_list)


def heisenberg_hamiltonian(edges: torch.Tensor) -> Hamiltonian:
    r"""Compute the Heisenberg Hamiltonian.

    .. math::

        \begin{align}
            H_{Heisenberg}= \sum_{(u,v) \in E(u>v)}(\alpha_{uv}X_u X_v + \beta_{uv}Y_u Y_v + \gamma_{uv}Z_u Z_v)
        \end{align}

    Args:
        edges: A tensor E shape=[3, V, V], where
               E[0][u][v] is \alpha_{uv},
               E[1][u][v] is \beta_{uv}, and
               E[2][u][v] is \gamma_{uv}.

    Returns:
        Hamiltonian representation of the Heisenberg Hamiltonian.

    Examples:
        .. code-block:: python

            import torch
            from quairkit.database.hamiltonian import heisenberg_hamiltonian

            edges = torch.tensor([
                [
                    [0, 0.5, 0],
                    [0.5, 0, 0.2],
                    [0, 0.2, 0]
                ],
                [
                    [0, 0.3, 0],
                    [0.3, 0, 0.4],
                    [0, 0.4, 0]
                ],
                [
                    [0, 0.7, 0],
                    [0.7, 0, 0.1],
                    [0, 0.1, 0]
                ]
            ])
            H_Heisenberg = heisenberg_hamiltonian(edges)
            print(f'The Heisenberg Hamiltonian is:\n{H_Heisenberg}')

        ::

            The Heisenberg Hamiltonian is:
            0.5 X0, X1
            0.30000001192092896 Y0, Y1
            0.699999988079071 Z0, Z1
            0.20000000298023224 X1, X2
            0.4000000059604645 Y1, Y2
            0.10000000149011612 Z1, Z2
    """
    h_list = []
    shape_of_edges = edges.shape

    assert len(shape_of_edges) == 3, \
        f'The input variable edges should be a 3-dimension torch.tensor, but receive {len(shape_of_edges)}-dimension'

    edges_param_list = edges.tolist()

    for i in range(shape_of_edges[1]):
        for j in range(i + 1, shape_of_edges[1]):
            alpha = edges_param_list[0][i][j]
            beta = edges_param_list[1][i][j]
            gamma = edges_param_list[2][i][j]
            if alpha == 0 and beta == 0 and gamma == 0:
                continue
            h_list.extend(
                (
                    (alpha, f'X{i},X{j}'),
                    (beta, f'Y{i},Y{j}'),
                    (gamma, f'Z{i},Z{j}')
                )
            )
    return Hamiltonian(h_list)



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Common functions for the State class.
"""

import warnings
from functools import reduce
from typing import Dict, List, Optional, Type, Union

import numpy as np
import torch

from .. import get_backend, get_dtype, utils
from . import MixedState, PureState, State

__state_dict: Dict[str, State] = {'state_vector': PureState, 'density_matrix': MixedState}


# TODO: should be default instead of choosing state_vector
def __state_convert(input_state: State, backend: Optional[str], system_dim: Union[int, List[int]]) -> State:
    r"""Copy and convert the input state to the expected state.
    
    Args:
        state: the input state.
        backend: the backend of the input state.
        system_dim: a list of dimensions for each system.
    
    Returns:
        state in the specified backend.
    
    """
    system_dim = [system_dim] * input_state.num_systems if isinstance(system_dim, int) else system_dim
    backend = input_state.backend if backend is None else backend
    input_state.reset_sequence()
    
    assert input_state.dim == int(np.prod(system_dim)), \
        f"The state dimension {input_state.dim} does not match with the input system dimension: {system_dim}."
    initializer = __state_dict[backend]
    return initializer(input_state.fit(backend), system_dim, list(range(len(system_dim))), 
                       [prob.clone() for prob in input_state._prob])


def __fetch_default_init(list_dim: List[int]) -> Type[State]:
    r"""Determine whether the input dimension should be a (batch of) pure/mixed state in ``'default'`` backend.
    
    Note:
        The convention is given by the table in ``to_state``.
    """
    if len(list_dim) == 1:
        return PureState

    if len(list_dim) == 2:
        return MixedState if list_dim[0] == list_dim[1] else PureState
    
    # batch case
    return MixedState if list_dim[-1] > 1 else PureState


def to_state(
        data: Union[torch.Tensor, np.ndarray, State], 
        system_dim: Union[int, List[int]] = 2,
        dtype: Optional[str] = None,
        state_backend: Optional[str] = None,
        eps: Optional[float] = 1e-4,
        prob: Optional[List[torch.Tensor]] = None
) -> Union[PureState, MixedState]:
    r"""The function to generate a specified state instance.

    Args:
        data: a representation of the quantum state in allowable backend, or an instance of the State class.
        system_dim: (list of) dimension(s) of the systems, can be a list of integers or an integer.
         For example, ``system_dim = 2`` means the systems are qubits (default setup); ``system_dim = [2, 3]`` means 
         the first system is a qubit and the second is a qutrit.
        dtype: Used to specify the data dtype of the data. Defaults to the global setup.
        state_backend: The backend of the state. Specified only when the input data is an instance of the State class.
        eps: The tolerance for checking the validity of the input state. Can be adjusted to ``None`` to disable the check.
        prob: The (list of) probability distribution of the state. The length of the list denotes the number of distributions.

    Returns:
        The generated quantum state.
    
    Note:
        When the ``backend`` is set as ``'default'``, the backend of this state is determined by the table
        
        +----------------+---------------------+---------------------+
        |                | single              | batch               |
        +================+=====================+=====================+
        | state_vector   | [d], [1, d], [d, 1] | [d1, ..., dn, d, 1] |
        +----------------+---------------------+---------------------+
        | density_matrix | [d, d]              | [d1, ..., dn, d, d] |
        +----------------+---------------------+---------------------+
    
    """
    if isinstance(data, State):
        assert (state_backend is not None) or (system_dim is not None), \
            "The backend or the system dimension that the input state converts to must be specified."
        return __state_convert(data, state_backend, system_dim)
    
    dtype = get_dtype() if dtype is None else dtype
    data = data.to(dtype=dtype) if isinstance(data, torch.Tensor) else \
        torch.from_numpy(data).to(dtype=dtype)
    list_dim = data.shape
    
    if get_backend() == 'default':
        initializer = __fetch_default_init(list_dim)
    else:
        raise NotImplementedError(
            f"the backend is not recognized or implemented: receive {get_backend()}")
    num_systems = initializer.check(data, system_dim, eps)
    system_dim = [system_dim] * num_systems if isinstance(system_dim, int) else system_dim
    return initializer(data, system_dim, list(range(num_systems)), prob)


def image_to_density_matrix(image_filepath: str) -> State:
    r"""Encode image to density matrix

    Args:
        image_filepath: Path to the image file.

    Returns:
        The density matrix obtained by encoding
    """
    import matplotlib
    image_matrix = matplotlib.image.imread(image_filepath)

    # Converting images to grayscale
    image_matrix = image_matrix.mean(axis=2)

    # Fill the matrix so that it becomes a matrix whose shape is [2**n,2**n]
    length = int(2 ** np.ceil(np.log2(np.max(image_matrix.shape))))
    image_matrix = np.pad(
        image_matrix,
        ((0, length - image_matrix.shape[0]), (0, length - image_matrix.shape[1])),
        "constant",
    )
    # Density matrix whose trace  is 1
    rho = image_matrix @ image_matrix.T
    rho = rho / np.trace(rho)
    return to_state(rho)


def __kron_state(state_1st: State, state_2nd: State) -> State:
    r"""Calculate the tensor product of two states.
    """
    system_dim = state_1st.system_dim + state_2nd.system_dim
    system_seq = state_1st.system_seq + [x + state_1st.num_systems for x in state_2nd.system_seq]
    
    if state_1st._prob:
        prob = state_1st._prob
        if state_2nd._prob:
            warnings.warn(
                "Detect tensor product of two probabilistic states: will discard prob info of the 2nd one", UserWarning)
    else:
        prob = state_2nd._prob
    
    if state_1st.backend == 'state_vector' and state_2nd.backend == 'state_vector':
        data = utils.linalg._kron(state_1st.ket, state_2nd.ket)
        return PureState(data, system_dim, system_seq, prob)
    else:
        data = utils.linalg._kron(state_1st.density_matrix, state_2nd.density_matrix)
        return MixedState(data, system_dim, system_seq, prob)


def tensor_state(state_1st: State, *args: State) -> State:
    r"""calculate tensor product (kronecker product) between at least two state. This function automatically returns State instance

    Args:
        state_1st: the first state
        args: other states

    Returns:
        tensor product state of input states

    Note:
        Need to be careful with the backend of states; Support broadcasting for batch states.
        Use ``quairkit.linalg.NKron`` if the input datatype is ``torch.Tensor`` or ``numpy.ndarray``.

    """
    return reduce(__kron_state, args, state_1st) if args else state_1st



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The library of random data generation functions
"""

import math
from typing import Iterable, List, Optional, Tuple, Union

import numpy as np
import scipy
import torch
from scipy.stats import unitary_group

# TODO this is added due to channel_repr_convert, move it to intrinsic
import quairkit as qkit

from ..core import Hamiltonian, State, get_dtype, get_float_dtype, to_state
from ..core.intrinsic import _alias, _format_total_dim
from ..core.utils.linalg import _dagger

__all__ = [
    "random_pauli_str_generator",
    "random_state",
    "random_hamiltonian_generator",
    "random_hermitian",
    "random_projector",
    "random_orthogonal_projection",
    "random_density_matrix",
    "random_unitary",
    "random_unitary_hermitian",
    "random_unitary_with_hermitian_block",
    "haar_orthogonal",
    "haar_unitary",
    "haar_state_vector",
    "haar_density_operator",
    "random_channel",
    "random_clifford"
]


def random_pauli_str_generator(num_qubits: int, terms: Optional[int] = 3) -> List:
    r"""Generate a random observable in list form.

    An observable :math:`O=0.3X\otimes I\otimes I+0.5Y\otimes I\otimes Z`'s list form is
    ``[[0.3, 'x0'], [0.5, 'y0,z2']]``. Such an observable is generated by
    ``random_pauli_str_generator(3, terms=2)``.

    Args:
        num_qubits: Number of qubits.
        terms: Number of terms in the observable. Defaults to 3.

    Returns:
        The Hamiltonian of randomly generated observable.

    Examples:
        .. code-block:: python

            observable = random_pauli_str_generator(num_qubits=2, terms=2)
            print(f'The Hamiltonian of randomly generated observable is:\n{observable}')

        ::

            The Hamiltonian of randomly generated observable is:
            [[-0.6019637631250563, 'y0'], [0.12777564473712655, 'x0,z1']]
    """
    pauli_str = []
    for sublen in np.random.randint(1, high=num_qubits + 1, size=terms):
        # Tips: -1 <= coeff < 1
        coeff = np.random.rand() * 2 - 1
        ops = np.random.choice(["x", "y", "z"], size=sublen)
        pos = np.random.choice(range(num_qubits), size=sublen, replace=False)
        op_list = [ops[i] + str(pos[i]) for i in range(sublen)]
        op_list.sort(key=lambda x: int(x[1:]))
        pauli_str.append([coeff, ",".join(op_list)])
    return pauli_str


@_alias({"num_systems": "num_qubits"})
def random_state(
    num_systems: int,
    rank: Optional[int] = None,
    is_real: Optional[bool] = False,
    size: Optional[Union[int, List[int]]] = 1,
    system_dim: Union[List[int], int] = 2,
) -> State:
    r"""Generate a random quantum state.

    Args:
        num_systems: The number of qubits contained in the quantum state.
        rank: The rank of the density matrix. Defaults to ``None`` (full rank).
        is_real: If the quantum state only contains real numbers. Defaults to ``False``.
        size: Batch size. Defaults to 1.
        system_dim: Dimension of systems. Can be a list of system dimensions
            or an int representing the dimension of all systems. Defaults to the qubit case.

    Raises:
        NotImplementedError: If the backend is wrong or not implemented.

    Returns:
        The generated quantum state.

    Examples:
        .. code-block:: python

            state = random_state(num_systems=1, size=1)
            print(f'The generated quantum state is:\n{state}')

            state = random_state(num_systems=1, rank=1)
            print(f'The generated quantum state is:\n{state}')

            state = random_state(num_systems=2, system_dim=[1, 2])
            print(f'The generated quantum state is:\n{state}')

            state = random_state(num_systems=3, size=[2, 1], system_dim=[1, 2, 1])
            print(f'The generated quantum state is:\n{state}')

        ::

            The generated quantum state is:
            ---------------------------------------------------
            Backend: state_vector
            System dimension: [2]
            System sequence: [0]
            [-0.33+0.72j -0.28+0.55j]
            ---------------------------------------------------
            The generated quantum state is:
            ---------------------------------------------------
            Backend: state_vector
            System dimension: [2]
            System sequence: [0]
            [ 0.83-0.31j -0.45+0.16j]
            ---------------------------------------------------
            The generated quantum state is:
            ---------------------------------------------------
            Backend: density_matrix
            System dimension: [1, 2]
            System sequence: [0, 1]
            [[ 0.63+0.j   -0.12-0.13j]
             [-0.12+0.13j  0.37-0.j  ]]
            ---------------------------------------------------
            The generated quantum state is:
            ---------------------------------------------------
            Backend: density_matrix
            System dimension: [1, 2, 1]
            System sequence: [0, 1, 2]
            Batch size: [2, 1]
            # 0:
            [[0.43+0.j   0.24-0.37j],
             [0.24+0.37j 0.57-0.j  ]]
            # 1:
            [[ 0.88+0.j   -0.29-0.11j],
             [-0.29+0.11j  0.12-0.j  ]]
            ---------------------------------------------------
    """
    dim = _format_total_dim(num_systems, system_dim)
    size = [size] if isinstance(size, int) else list(size)
    rank = np.random.randint(1, dim + 1) if rank is None else rank
    total_size = int(np.prod(size))

    if rank == 1:
        list_state = torch.stack(
            [haar_state_vector(dim, is_real) for _ in range(total_size)]
        ).view(size + [dim, 1])
    else:
        list_state = torch.stack(
            [haar_density_operator(dim, rank, is_real) for _ in range(total_size)]
        ).view(size + [dim, dim])

    list_state = list_state if total_size > 1 else list_state.squeeze()
    return to_state(list_state, system_dim)


def random_hamiltonian_generator(
    num_qubits: int, terms: Optional[int] = 3
) -> Hamiltonian:
    r"""Generate a random Hamiltonian.

    Args:
        num_qubits: Number of qubits.
        terms: Number of terms in the Hamiltonian. Defaults to 3.

    Returns:
        The randomly generated Hamiltonian.

    Examples:
        .. code-block:: python

            Hamiltonian = random_hamiltonian_generator(3, 2)
            print(f'The randomly generated Hamiltonian is:\n{Hamiltonian}')

        ::

            The randomly generated Hamiltonian is:
            0.11801365595625102 Z0, X2
            0.9059897222160238 X0, Z1, Z2
    """
    return Hamiltonian(random_pauli_str_generator(num_qubits, terms))


def random_hermitian(num_qubits: int) -> torch.Tensor:
    r"""Randomly generate a normalized :math:`2^n \times 2^n` Hermitian matrix.

    Args:
        num_qubits: Number of qubits :math:`n`.

    Returns:
         A :math:`2^n \times 2^n` Hermitian matrix.

    Examples:
        .. code-block:: python

            Hermitian = random_hermitian(1)
            print(f'The randomly generated Hermitian matrix is:\n{Hermitian}')

        ::

            The randomly generated Hermitian matrix is:
            tensor([[0.1038+0.0000j, 0.4251+0.2414j],
                    [0.4251-0.2414j, 0.7333+0.0000j]])
    """
    assert num_qubits > 0
    n = 2**num_qubits

    mat = np.random.randn(n, n) + 1j * np.random.randn(n, n)
    for i in range(n):
        mat[i, i] = np.abs(mat[i, i])
        for j in range(i):
            mat[i, j] = np.conj(mat[j, i])

    eigval = np.linalg.eigvalsh(mat)
    max_eigval = np.max(np.abs(eigval))
    return torch.tensor(mat / max_eigval, dtype=get_dtype())


@_alias({"num_systems": "num_qubits"})
def random_projector(num_systems: int, size: int = 1,
                     system_dim: Union[List[int], int] = 2) -> torch.Tensor:
    r"""Randomly generate a :math:`d \times d` orthogonal projector.

    Args:
        num_systems: Number of systems.
        size: Number of projectors.
        system_dim: Dimension of systems. Can be a list of system dimensions 
            or an int representing the dimension of all systems. Defaults to the qubit case.

    Returns:
         A :math:`2^n \times 2^n` orthogonal projector.

    Examples:
        .. code-block:: python

            Orthogonal_projector = random_orthogonal_projection(1)
            print(f'The randomly generated Orthogonal_projector is:\n{Orthogonal_projector}')

        ::

            The randomly generated Orthogonal_projector is:
            tensor([[0.9704+1.1123e-10j, 0.1692+7.7007e-03j],
                    [0.1692-7.7008e-03j, 0.0296-1.1123e-10j]])
    """
    U = random_unitary(num_systems, size, system_dim)
    rank = np.random.randint(1, U.shape[0])
    V = U[:, :rank]
    return V @ _dagger(V)


random_orthogonal_projection = random_projector


def random_density_matrix(num_qubits: int) -> torch.Tensor:
    r"""Randomly generate an num_qubits-qubit state in density matrix form.

    Args:
        num_qubits: Number of qubits :math:`n`.

    Returns:
        A :math:`2^n \times 2^n` density matrix.

    Examples:
        .. code-block:: python

            Density_matrix = random_density_matrix(1)
            print(f'The randomly generated density matrix is:\n{Density_matrix}')

        ::

            The randomly generated density matrix is:
            tensor([[0.3380+0.0000j, 0.4579+0.1185j],
                    [0.4579-0.1185j, 0.6620+0.0000j]])
    """
    dim = 2**num_qubits
    return haar_density_operator(dim, rank=np.random.randint(1, dim + 1))


@_alias({"num_systems": "num_qubits"})
def random_unitary(
    num_systems: int,
    size: Union[int, List[int]] = 1,
    system_dim: Union[List[int], int] = 2,
) -> torch.Tensor:
    r"""Randomly generate a :math:`d \times d` unitary.

    Args:
        num_systems: Number of systems in this unitary (alias of ``num_qubits``).
        size: Batch size. Defaults to 1.
        system_dim: Dimension of systems. Can be a list of system dimensions or an int representing
            the dimension of all systems. Defaults to the qubit case.

    Returns:
         A :math:`d \times d` unitary matrix.

    Examples:
        .. code-block:: python

            unitary_matrix_1 = random_unitary(num_systems=1, system_dim=2)
            print(f'The randomly generated unitary_matrix_1 is:\n{unitary_matrix_1}')

            unitary_matrix_2 = random_unitary(num_systems=2, system_dim=[1, 2])
            print(f'The randomly generated unitary_matrix_2 is:\n{unitary_matrix_2}')

            unitary_matrix_3 = random_unitary(num_systems=1, size=[1,2])
            print(f'The randomly generated unitary_matrix_3 is:\n{unitary_matrix_3}')

        ::

            The randomly generated unitary_matrix_1 is:
            tensor([[-0.5288+0.5296j, -0.5277-0.4019j],
                    [-0.5321-0.3959j, -0.3627+0.6546j]])
            The randomly generated unitary_matrix_2 is:
            tensor([[0.6996-0.1504j, 0.3414+0.6095j],
                    [0.4954-0.4925j, -0.6315-0.3364j]])
            The randomly generated unitary_matrix_3 is:
            tensor([[[[0.3240+0.1404j, 0.4166-0.8377j],
                      [0.6068+0.7121j, -0.2804+0.2146j]],

                     [[-0.2620-0.0886j, -0.3587+0.8916j],
                      [0.5196-0.8084j, 0.2238+0.1624j]]]])
    """
    dim = _format_total_dim(num_systems, system_dim)
    size = [size] if isinstance(size, int) else list(size)
    total_size = math.prod(size)
    list_unitary = torch.stack(
        [
            torch.tensor(unitary_group.rvs(dim), dtype=get_dtype())
            for _ in range(total_size)
        ]
    ).view(size + [dim, dim])
    return list_unitary if total_size > 1 else list_unitary.squeeze()


def random_unitary_hermitian(num_qubits: int) -> torch.Tensor:
    r"""Randomly generate a :math:`2^n \times 2^n` Hermitian unitary.

    Args:
        num_qubits: Number of qubits :math:`n`.

    Returns:
         A :math:`2^n \times 2^n` Hermitian unitary matrix.

    Examples:
        .. code-block:: python

            Unitary_hermitian = random_unitary_hermitian(1)
            print(f'The randomly generated hermitian unitary is:\n{Unitary_hermitian}')

        ::

            The randomly generated hermitian unitary is:
            tensor([[0.2298+2.2018e-09j, -0.8408+4.9013e-01j],
                    [-0.8408-4.9013e-01j, -0.2298-2.2018e-09j]])
    """
    proj_mat = random_orthogonal_projection(num_qubits)
    id_mat = torch.eye(2**num_qubits)
    return (2 + 0j) * proj_mat - id_mat


def random_unitary_with_hermitian_block(
    num_qubits: int, is_unitary: bool = False
) -> torch.Tensor:
    r"""Randomly generate a unitary :math:`2^n \times 2^n` matrix that is a block encoding of a :math:`2^{n/2} \times 2^{n/2}` Hermitian matrix.

    Args:
        num_qubits: Number of qubits :math:`n`.
        is_unitary: Whether the Hermitian block is a unitary divided by 2 (for tutorial only).

    Returns:
         A :math:`2^n \times 2^n` unitary matrix whose upper-left block is Hermitian.

    Examples:
        .. code-block:: python

            unitary_matrix_1 = random_unitary_with_hermitian_block(num_qubits=2, is_unitary=False)
            print(f'The randomly generated unitary matrix 1 with hermitian block is:\n{unitary_matrix_1}')

            unitary_matrix_2 = random_unitary_with_hermitian_block(num_qubits=2, is_unitary=True)
            print(f'The randomly generated unitary matrix 2 with hermitian block is:\n{unitary_matrix_2}')

        ::

            The randomly generated unitary matrix 1 with hermitian block is:
            tensor([[ 5.7873e-01+0.0000j,  2.2460e-01+0.2711j, -1.5514e-08+0.5646j,
                      3.6316e-01-0.3009j],
                    [ 2.2460e-01-0.2711j,  7.0578e-01+0.0000j, -3.6316e-01-0.3009j,
                     -2.2489e-08+0.3944j],
                    [-1.5514e-08+0.5646j,  3.6316e-01-0.3009j,  5.7873e-01+0.0000j,
                      2.2460e-01+0.2711j],
                    [-3.6316e-01-0.3009j, -2.2489e-08+0.3944j,  2.2460e-01-0.2711j,
                      7.0578e-01+0.0000j]])
            The randomly generated unitary matrix 2 with hermitian block is:
            tensor([[-1.8185e-01-1.6847e-09j,  3.0894e-01+3.4855e-01j,
                     2.2516e-09+8.6603e-01j, -1.4451e-09-9.3500e-11j],
                    [ 3.0894e-01-3.4855e-01j,  1.8185e-01+1.6847e-09j,
                     1.7456e-09-9.3500e-11j,  2.4038e-09+8.6603e-01j],
                    [ 2.2516e-09+8.6603e-01j, -1.4451e-09-9.3500e-11j,
                    -1.8185e-01-1.6847e-09j,  3.0894e-01+3.4855e-01j],
                    [ 1.7456e-09-9.3500e-11j,  2.4038e-09+8.6603e-01j,
                     3.0894e-01-3.4855e-01j,  1.8185e-01+1.6847e-09j]])
    """
    assert num_qubits > 0

    if is_unitary:
        mat0 = random_unitary_hermitian(num_qubits - 1).detach().numpy() / 2
    else:
        mat0 = random_hermitian(num_qubits - 1).detach().numpy()
    id_mat = np.eye(2 ** (num_qubits - 1))
    mat1 = 1j * scipy.linalg.sqrtm(id_mat - np.matmul(mat0, mat0))

    mat = np.block([[mat0, mat1], [mat1, mat0]]).astype(complex)

    return torch.tensor(mat, dtype=get_dtype())


def haar_orthogonal(dim: int) -> torch.Tensor:
    r"""Randomly generate an orthogonal matrix following Haar measure,
    referenced by arXiv:math-ph/0609050v2.

    Args:
        dim: Dimension of the orthogonal matrix.

    Returns:
        A :math:`d \times d` orthogonal matrix.

    Examples:
        .. code-block:: python

            Haar_orthogonal = haar_orthogonal(2)
            print(f'The randomly generated orthogonal matrix is:\n{Haar_orthogonal}')

        ::

            The randomly generated orthogonal matrix is:
            tensor([[-0.6859+0.j,  0.7277+0.j],
                    [ 0.7277+0.j,  0.6859+0.j]])
    """
    # Step 1: sample from Ginibre ensemble
    ginibre = np.random.randn(dim, dim)
    # Step 2: perform QR decomposition of G
    mat_q, mat_r = np.linalg.qr(ginibre)
    # Step 3: make the decomposition unique
    mat_lambda = np.diag(mat_r) / abs(np.diag(mat_r))
    mat_u = mat_q @ np.diag(mat_lambda)
    return torch.tensor(mat_u, dtype=get_dtype())


def haar_unitary(dim: int) -> torch.Tensor:
    r"""Randomly generate a unitary following Haar measure,
    referenced by arXiv:math-ph/0609050v2.

    Args:
        dim: Dimension of the unitary.

    Returns:
        A :math:`d \times d` unitary.

    Examples:
        .. code-block:: python

            Haar_unitary = haar_unitary(2)
            print(f'The randomly generated unitary is:\n{Haar_unitary}')

        ::

            The randomly generated unitary is:
            tensor([[ 0.2800+0.6235j,  0.7298+0.0160j],
                    [-0.7289-0.0396j,  0.3267-0.6003j]])
    """
    # Step 1: sample from Ginibre ensemble
    ginibre = (np.random.randn(dim, dim) + 1j * np.random.randn(dim, dim)) / np.sqrt(2)
    # Step 2: perform QR decomposition of G
    mat_q, mat_r = np.linalg.qr(ginibre)
    # Step 3: make the decomposition unique
    mat_lambda = np.diag(mat_r) / np.abs(np.diag(mat_r))
    mat_u = mat_q @ np.diag(mat_lambda)
    return torch.tensor(mat_u, dtype=get_dtype())


def haar_state_vector(dim: int, is_real: Optional[bool] = False) -> torch.Tensor:
    r"""Randomly generate a state vector following Haar measure.

    Args:
        dim: Dimension of the state vector.
        is_real: Whether the vector is real. Defaults to False.

    Returns:
        A :math:`d \times 1` state vector.

    Examples:
        .. code-block:: python

            Haar_state_vector = haar_state_vector(3, is_real=True)
            print(f'The randomly generated state vector is:\n{Haar_state_vector}')

        ::

            The randomly generated state vector is:
            tensor([[0.9908+0.j],
                    [-0.1356+0.j]])
    """
    if is_real:
        # Generate a Haar random orthogonal matrix
        mat_orthog = haar_orthogonal(dim)
        # Perform orthogonal transformation on |0>
        phi = mat_orthog[:, 0]
    else:
        # Generate a Haar random unitary
        unitary = haar_unitary(dim)
        # Perform unitary transformation on |0>
        phi = unitary[:, 0]

    return phi.view([-1, 1])


def haar_density_operator(
    dim: int, rank: int, is_real: Optional[bool] = False
) -> torch.Tensor:
    r"""Randomly generate a density matrix following Haar measure.

    Args:
        dim: Dimension of the density matrix.
        rank: Rank of the density matrix.
        is_real: Whether the density matrix is real. Defaults to False.

    Returns:
        A :math:`d \times d` density matrix.

    Examples:
        .. code-block:: python

            rho1 = haar_density_operator(dim=2, rank=2)
            print(f'The randomly generated density matrix 1 is:\n{rho1}')

            rho2 = haar_density_operator(dim=2, rank=1, is_real=True)
            print(f'The randomly generated density matrix 2 is:\n{rho2}')

        ::

            The randomly generated density matrix 1 is:
            tensor([[0.8296+1.1215e-18j, -0.0430+3.5193e-01j],
                    [-0.0430-3.5193e-01j, 0.1704-1.1215e-18j]])
            The randomly generated density matrix 2 is:
            tensor([[0.6113+0.j, -0.4875+0.j],
                    [-0.4875+0.j, 0.3887+0.j]])
    """
    assert 0 < rank <= dim, "rank is an invalid number"
    if is_real:
        ginibre_matrix = np.random.randn(dim, rank)
        rho = ginibre_matrix @ ginibre_matrix.T
    else:
        ginibre_matrix = np.random.randn(dim, rank) + 1j * np.random.randn(dim, rank)
        rho = ginibre_matrix @ ginibre_matrix.conj().T
    rho = rho / np.trace(rho)
    return torch.tensor(rho, dtype=get_dtype())


@_alias({"num_systems": "num_qubits"})
def random_channel(
    num_systems: int,
    rank: int = None,
    target: str = "kraus",
    size: Optional[int] = 1,
    system_dim: Union[List[int], int] = 2,
) -> torch.Tensor:
    r"""Generate a random channel from its Stinespring representation.

    Args:
        num_systems: Number of systems.
        rank: Rank of this channel. If None, it is randomly sampled from :math:`[1, d]`.
        target: Target representation; should be ``'choi'``, ``'kraus'`` or ``'stinespring'``.
        size: Batch size. Defaults to 1.
        system_dim: Dimension of systems. Can be a list of system dimensions or an int representing
            the dimension of all systems. Defaults to the qubit case.

    Returns:
        The target representation of a random channel.
        
    Examples:
        .. code-block:: python

            channel_kraus = random_channel(num_systems=1, target="kraus", system_dim=2)
            print(f'The randomly generated kraus channel is:\n{channel_kraus}')

            channel_choi = random_channel(num_systems=1, target="choi", system_dim=2)
            print(f'The randomly generated choi channel is:\n{channel_choi}')

            channel_stinespring = random_channel(num_systems=1, target="stinespring", size=1, rank=1)
            print(f'The randomly generated stinespring channel is:\n{channel_stinespring}')

            batch_channels = random_channel(num_systems=2, size=2, target="kraus", system_dim=[1,2], rank=2)
            print(f'The randomly generated kraus channel is:\n{batch_channels}')

        ::

            The randomly generated kraus channel is:
            tensor([[[ 0.2361+0.7497j,  0.0224+0.6178j],
                     [ 0.5942-0.1706j, -0.7860+0.0085j]]])
            The randomly generated choi channel is:
            tensor([[ 0.5136+0.0000j, -0.4784-0.1446j, -0.2850-0.4106j, -0.1583-0.4886j],
                    [-0.4784+0.1446j,  0.4864+0.0000j,  0.3811+0.3023j,  0.2850+0.4106j],
                    [-0.2850+0.4106j,  0.3811-0.3023j,  0.4864+0.0000j,  0.4784+0.1446j],
                    [-0.1583+0.4886j,  0.2850-0.4106j,  0.4784-0.1446j,  0.5136+0.0000j]])
            The randomly generated stinespring channel is:
            tensor([[ 0.1652+0.5347j,  0.6310+0.5372j],
                    [ 0.4751+0.6790j, -0.5167-0.2150j]])
            The randomly generated kraus channel is:
            tensor([[[[-0.3189-0.6385j, -0.1873+0.1634j],
                      [ 0.2218+0.0440j, -0.6294+0.0947j]],

                     [[-0.0072+0.0749j, -0.0790+0.6740j],
                      [-0.5121-0.4142j, -0.0357-0.2671j]]],


                    [[[-0.1163-0.1931j,  0.2001-0.5852j],
                      [-0.5174+0.5253j,  0.1128-0.2459j]],

                     [[-0.1362+0.3280j, -0.2677+0.5444j],
                      [ 0.4328-0.3033j, -0.3862-0.1646j]]]])
    """
    target = target.lower()
    dim = _format_total_dim(num_systems, system_dim)
    rank = np.random.randint(dim) + 1 if rank is None else rank
    assert (
        1 <= rank <= dim
    ), f"rank must be positive and no larger than the dimension {dim} of the channel: received {rank}"

    list_repr = []

    for _ in range(size):

        unitary = unitary_group.rvs(rank * dim)
        stinespring_mat = torch.tensor(unitary[:, :dim], dtype=get_dtype()).reshape(
            [rank, dim, dim]
        )
        list_kraus = stinespring_mat[:rank]

        if target == "choi":
            list_repr.append(
                qkit.qinfo.channel_repr_convert(
                    list_kraus, source="kraus", target="choi"
                )
            )
        elif target == "stinespring":
            list_repr.append(
                qkit.qinfo.channel_repr_convert(
                    list_kraus, source="kraus", target="stinespring"
                )
            )
        else:
            list_repr.append(list_kraus)

    return torch.stack(list_repr) if size > 1 else list_repr[0]


def random_clifford(num_qubits: int) -> torch.Tensor:
    r"""Generate a random Clifford unitary.

    Args:
        num_qubits: The number of qubits (n).

    Returns:
        The matrix form of a random Clifford unitary.

    Reference:
        Sergey Bravyi and Dmitri Maslov, *Hadamard-free circuits expose the structure of the Clifford group*. 
        IEEE Transactions on Information Theory 67(7), 4546-4563 (2021).

    Examples:
        .. code-block:: python

            num_qubits = 1
            clifford_matrix = random_clifford(num_qubits)
            print(f'The randomly generated Clifford unitary is:\n{clifford_matrix}')

        ::

            The randomly generated Clifford unitary is:
            tensor([[ 0.7071+0.0000j, -0.7071+0.0000j],
                    [ 0.0000+0.7071j,  0.0000+0.7071j]])
    """
    # Generate a random Clifford operator (stabilizer tableaux) and the elements of the canonical form
    _, gamma_matrices, delta_matrices, hadamard_layer, permutation = __generate_random_clifford(num_qubits)

    gamma1, gamma2 = gamma_matrices[0], gamma_matrices[1]
    delta1, delta2 = delta_matrices[0].T, delta_matrices[1].T
    phase = np.random.randint(0, 2, size=2 * num_qubits)

    # Initialize the circuit
    circuit = qkit.Circuit(num_qubits)

    __apply_circuit_layers(circuit, delta2, gamma2)

    # Apply Pauli gates based on the phase vector
    x_indices, y_indices, z_indices = [], [], []
    for idx in range(num_qubits):
        if phase[idx] == 1 and phase[idx + num_qubits] == 0:
            x_indices.append(idx)
        elif phase[idx] == 0 and phase[idx + num_qubits] == 1:
            y_indices.append(idx)
        elif phase[idx] == 1 and phase[idx + num_qubits] == 1:
            z_indices.append(idx)
    if x_indices:
        circuit.x(x_indices)
    if y_indices:
        circuit.y(y_indices)
    if z_indices:
        circuit.z(z_indices)

    # Apply SWAP gates based on the permutation
    swapped_indices, swap_gate_indices = [], []
    for idx in range(num_qubits):
        if permutation[idx] == idx:
            continue
        swapped_indices.append(permutation[idx])
        if idx in swapped_indices:
            continue
        swap_gate_indices.append([idx, permutation[idx]])
    if swap_gate_indices:
        circuit.swap(swap_gate_indices)

    if hadamard_gate_indices := np.argwhere(hadamard_layer == 1).tolist():
        circuit.h(hadamard_gate_indices)

    __apply_circuit_layers(circuit, delta1, gamma1)

    # Return the Clifford unitary matrix
    return circuit.matrix


def __apply_circuit_layers(circuit, delta: Iterable[int], gamma: Iterable[int]) -> None:
    r"""Apply CNOT, CZ, and S gates to the circuit based on Delta and Gamma matrices."""
    if cnot_gate_indices := np.argwhere(np.triu(delta, k=1) == 1).tolist():
        circuit.cnot(cnot_gate_indices)
    if cz_gate_indices := np.argwhere(np.triu(gamma, k=1) == 1).tolist():
        circuit.cz(cz_gate_indices)
    if s_gate_indices := np.argwhere(np.diag(gamma) == 1).tolist():
        circuit.s(s_gate_indices)


def __generate_random_clifford(
    num_qubits: int,
) -> Tuple[np.ndarray, List[np.ndarray], List[np.ndarray], np.ndarray]:
    r"""Generate a random Clifford operator and its canonical form elements.

    TODO: bad code quality, need to eliminate the for loops.
    """
    # Constant matrices
    zero_matrix = np.zeros((num_qubits, num_qubits), dtype=int)
    zero_matrix_2n = np.zeros((2 * num_qubits, 2 * num_qubits), dtype=int)
    identity_matrix = np.eye(num_qubits, dtype=int)

    # Sample from the quantum Mallows distribution
    hadamard_layer, permutation = __sample_quantum_mallows(num_qubits)

    gamma1 = np.copy(zero_matrix)
    delta1 = np.copy(identity_matrix)
    gamma2 = np.copy(zero_matrix)
    delta2 = np.copy(identity_matrix)

    # Generate random diagonal elements for Gamma matrices
    for i in range(num_qubits):
        gamma2[i, i] = np.random.randint(2)
        if hadamard_layer[i]:
            gamma1[i, i] = np.random.randint(2)

    # Generate random elements for Gamma and Delta matrices based on canonical form constraints
    for j in range(num_qubits):
        for i in range(j + 1, num_qubits):
            b = np.random.randint(2)
            gamma2[i, j] = b
            gamma2[j, i] = b
            delta2[i, j] = np.random.randint(2)
            if hadamard_layer[i] == 1 and hadamard_layer[j] == 1:
                b = np.random.randint(2)
                gamma1[i, j] = b
                gamma1[j, i] = b
            if hadamard_layer[i] == 1 and hadamard_layer[j] == 0 and permutation[i] < permutation[j]:
                b = np.random.randint(2)
                gamma1[i, j] = b
                gamma1[j, i] = b
            if hadamard_layer[i] == 0 and hadamard_layer[j] == 1 and permutation[i] > permutation[j]:
                b = np.random.randint(2)
                gamma1[i, j] = b
                gamma1[j, i] = b
            if hadamard_layer[i] == 0 and hadamard_layer[j] == 1:
                delta1[i, j] = np.random.randint(2)
            if hadamard_layer[i] == 1 and hadamard_layer[j] == 1 and permutation[i] > permutation[j]:
                delta1[i, j] = np.random.randint(2)
            if hadamard_layer[i] == 0 and hadamard_layer[j] == 0 and permutation[i] < permutation[j]:
                delta1[i, j] = np.random.randint(2)

    # Compute stabilizer tableaux
    prod1 = np.matmul(gamma1, delta1)
    prod2 = np.matmul(gamma2, delta2)
    inv1 = np.linalg.inv(np.transpose(delta1))
    inv2 = np.linalg.inv(np.transpose(delta2))
    f1 = np.block([[delta1, zero_matrix], [prod1, inv1]])
    f2 = np.block([[delta2, zero_matrix], [prod2, inv2]])
    f1 = f1.astype(int) % 2
    f2 = f2.astype(int) % 2

    # Compute the full stabilizer tableaux
    stabilizer_tableaux = np.copy(zero_matrix_2n)
    # Apply qubit permutation to F2
    for i in range(num_qubits):
        stabilizer_tableaux[i, :] = f2[permutation[i], :]
        stabilizer_tableaux[i + num_qubits, :] = f2[permutation[i] + num_qubits, :]
    # Apply Hadamard layer
    for i in range(num_qubits):
        if hadamard_layer[i] == 1:
            stabilizer_tableaux[(i, i + num_qubits), :] = stabilizer_tableaux[(i + num_qubits, i), :]

    gamma_matrices = [gamma1, gamma2]
    delta_matrices = [delta1, delta2]

    return np.matmul(f1, stabilizer_tableaux) % 2, gamma_matrices, delta_matrices, hadamard_layer, permutation


def __sample_quantum_mallows(n: int) -> Tuple[np.ndarray, np.ndarray]:
    r"""Sample from the quantum Mallows distribution.

    Returns:
        hadamard_layer: A binary array indicating the Hadamard layer.
        permutation: A permutation array.
    """
    # Initialize Hadamard layer and permutation layer
    hadamard_layer = np.zeros(n, dtype=int)
    permutation = np.zeros(n, dtype=int)

    remaining_indices = list(range(n))

    for i in range(n):
        m = n - i  # Number of remaining indices
        r = np.random.uniform(0, 1)
        index = -1 * int(np.ceil(np.log2(r + (1 - r) * (4 ** (-1 * float(m))))))  # sample index
        hadamard_layer[i] = 1 if index < m else 0
        k = index if index < m else 2 * m - index - 1
        permutation[i] = remaining_indices[k]
        del remaining_indices[k]

    return hadamard_layer, permutation



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


r"""
Check functions in QuAIRKit.
"""

import math
from typing import Callable, List, Optional, Tuple, Union

import numpy as np
import torch

from .. import utils


def _is_square(mat: torch.Tensor) -> bool:
    return mat.ndim >= 2 and mat.shape[-1] == mat.shape[-2]


def _is_hermitian(mat: torch.Tensor, eps: float = 1e-6) -> torch.Tensor:
    mat = mat.to(torch.complex128)
    return (mat - utils.linalg._dagger(mat)).norm(dim=(-2, -1)) < eps


def _is_positive(mat: torch.Tensor, eps: float = 1e-6) -> torch.Tensor:
    mat = mat.to(torch.complex128)
    
    herm_check = _is_hermitian(mat, eps)
    pos_check = torch.min(torch.linalg.eigvalsh(mat), dim=-1).values >= -eps
    
    return herm_check & pos_check


def _is_state_vector(vec: torch.Tensor, eps: float = 1e-6) -> torch.Tensor:
    vec = vec.to(torch.complex128)
    eps = min(eps * vec.shape[-2], 1e-4)
    
    vec_bra = utils.linalg._dagger(vec)
    return (vec_bra @ vec - (1 + 0j)).norm(dim=(-2, -1)) < eps


def _is_density_matrix(rho: torch.Tensor, eps: float = 1e-6) -> torch.Tensor:
    rho = rho.to(torch.complex128)
    eps = min(eps * rho.shape[-1], 1e-4)
    
    is_trace_one = torch.abs(utils.linalg._trace(rho, -2, -1) - 1) < eps
    is_pos = _is_positive(rho, eps)
    return is_trace_one & is_pos


def _is_projector(mat: torch.Tensor, eps: float = 1e-6) -> torch.Tensor:
    mat = mat.to(torch.complex128)
    return (mat @ mat - mat).norm(dim=(-2, -1)) < eps


def _is_isometry(mat: torch.Tensor, eps: float = 1e-6) -> torch.Tensor:
    mat = mat.to(torch.complex128)
    dim = mat.shape[-1]
    eps = min(eps * dim, 1e-2)
    
    identity = torch.eye(dim, device=mat.device)
    return (utils.linalg._dagger(mat) @ mat - identity).norm(dim=(-2, -1)) < eps


def _is_unitary(mat: torch.Tensor, eps: float = 1e-4) -> torch.Tensor:
    return _is_isometry(mat, eps) & _is_hermitian(utils.linalg._dagger(mat) @ mat, eps)


def _is_ppt(density_op: torch.Tensor, eps: float = 1e-6) -> torch.Tensor:
    density_op = density_op.to(torch.complex128)
    return utils.qinfo._negativity(density_op) <= eps


def _is_choi(op: torch.Tensor) -> torch.Tensor:
    op = op.to(torch.complex128)
    sys_dim = math.isqrt(op.shape[-1])
    
    is_pos = _is_positive(op)
    
    partial_op = utils.linalg._partial_trace(op, [1], [sys_dim, sys_dim])
    is_trace_non_inc = _is_positive(torch.eye(sys_dim).expand_as(partial_op) - partial_op)
    
    return is_pos & is_trace_non_inc


def _is_linear(
    func: Callable[[torch.Tensor], torch.Tensor],
    generator: Union[List[int], Callable[[], torch.Tensor]],
    input_dtype: torch.dtype,
    eps: float = 1e-5,
) -> torch.Tensor:
    list_err = []
    for _ in range(5):
        A = generator()
        B = generator()
        k = torch.rand(1, dtype=input_dtype)

        list_err.append(
            torch.norm(func(k * A + B) - k * func(A) - func(B)).view([1])
        )

    return torch.mean(torch.concat(list_err)) < eps


def _is_povm(set_op: torch.Tensor, eps: float = 1e-6) -> torch.Tensor:
    dim, batch_shape, set_op = set_op.shape[-1], list(set_op.shape[:-3]), set_op.view([-1] + list(set_op.shape[-3:]))
    
    pos_check = _is_positive(set_op.view([-1, dim, dim]), eps)
    pos_check = torch.all(pos_check.view(set_op.shape[:-2]), dim=-1)
    
    oper_sum = set_op.sum(dim=-3)
    identity = torch.eye(dim, device=oper_sum.device)
    complete_check = (identity - oper_sum).norm(dim=(-2, -1)) < eps
    return (pos_check & complete_check).view(batch_shape)


def _is_pvm(set_op: torch.Tensor, eps: float = 1e-6) -> torch.Tensor:
    set_op = set_op.to(torch.complex128)
    
    povm_check = _is_povm(set_op, eps)

    proj_check = _is_projector(set_op, eps)
    proj_check = torch.all(proj_check, dim=-1)
    
    set_isometry = set_op.view(list(set_op.shape[:-2]) + [-1])
    identity = torch.eye(set_isometry.shape[-2], device=set_isometry.device)
    isometry_check = (set_isometry @ utils.linalg._dagger(set_isometry) - identity).norm(dim=(-2, -1)) < eps

    return povm_check & proj_check & isometry_check



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The basic function of the QuAIRKit.
"""

import enum
import random
import warnings
from typing import Optional, Union

import numpy as np
import torch


class Backend(enum.Enum):
    r"""Backend classifier in the QuAIRKit.
    """
    Simulator = 'default'
    QPU = 'qpu' # TODO: to be added

    StateVector = 'state_vector' # TODO: depreciated in the future
    DensityMatrix = 'density_matrix' # TODO: depreciated in the future


DEFAULT_DEVICE = "cpu"
DEFAULT_SIMULATOR = Backend.Simulator
DEFAULT_DTYPE = torch.complex64
SEED = None


def set_device(device: "str") -> None:
    r"""Set the device to save the tensor.

    Args:
        device: The name of the device.
    """
    global DEFAULT_DEVICE
    DEFAULT_DEVICE = device
    torch.set_default_device(device)


def get_device() -> "str":
    r"""Get the current device to save the tensor.

    Returns:
        The name of the current device.
    """
    return DEFAULT_DEVICE


def set_backend(backend: Union[str, Backend]) -> None:
    r"""Set the backend implementation of QuAIRKit.

    Args:
        backend: The name of the backend.
    """
    global DEFAULT_SIMULATOR
    if backend in ["state_vector", "density_matrix"]:
        warnings.warn(
            "The usage for 'state_vector' and 'density_matrix' will be deprecated in the future."
            + "Please use 'default' instead.",
            DeprecationWarning,
        )
        backend = 'default'
    DEFAULT_SIMULATOR = Backend(backend) if isinstance(backend, str) else backend


def get_backend() -> str:
    r"""Get the current backend of QuAIRKit.

    Returns:
        The name of currently used backend.
    """
    return DEFAULT_SIMULATOR.value


def set_seed(seed: int) -> None:
    r"""Set the global seed of QuAIRKit.

    Args:
        seed: the random seed used in QuAIRKit.

    Note:
        The seed is set for the following modules:
        ``torch``, ``torch.cuda``, ``numpy``, ``random``

    """
    global SEED
    SEED = seed
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)


def get_seed() -> None:
    r"""Get the currently used seed of QuAIRKit.

    Returns:
        The Currently used seed.

    """
    return SEED


def set_dtype(dtype: str) -> None:
    r"""Set the data type .

    Args:
        dtype: The dtype can be ``complex64`` and ``complex128``.

    Raises:
        ValueError: The dtype should be complex64 or complex128.
    """
    global DEFAULT_DTYPE

    if dtype == "complex64":
        dtype = torch.complex64
    elif dtype == "complex128":
        dtype = torch.complex128
    else:
        raise ValueError("The dtype should be 'complex64' or 'complex128'.")
    DEFAULT_DTYPE = dtype

    float_dtype = torch.float32 if dtype == torch.complex64 else torch.float64
    torch.set_default_dtype(float_dtype)


def get_dtype() -> torch.dtype:
    r"""Return currently used data type.

    Returns:
        Currently used data type.
    """
    return DEFAULT_DTYPE


def get_float_dtype() -> torch.dtype:
    r"""Return currently used float data type.

    Returns:
        Currently used data type.
    """
    return torch.float32 if DEFAULT_DTYPE == torch.complex64 else torch.float64



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The source file of the basic class for the quantum channels.
"""

import math
import warnings
from typing import Iterable, List, Optional, Union

import torch

from ...core import MixedState, Operator, OperatorInfoType, State, utils
from ...core.intrinsic import _format_operator_idx
from ...core.utils.qinfo import (_choi_to_kraus, _choi_to_stinespring,
                                 _kraus_to_choi, _kraus_to_stinespring,
                                 _stinespring_to_choi, _stinespring_to_kraus)


class Channel(Operator):
    r"""Basic class for quantum channels.

    Args:
        type_repr: type of a representation, should be ``'choi'``, ``'kraus'``, ``'stinespring'`` or ``'gate'``.
        representation: the representation of this channel. Defaults to ``None`` i.e. not specified.
        system_idx: indices of the system on which this channel acts on. Defaults to ``None``.
            i.e. list(range(number of acted systems)).
        acted_system_dim: dimension of systems that this channel acts on. Can be a list of system dimensions 
            or an int representing the dimension of all systems. Defaults to ``None``.
        check_legality: whether check the completeness of the representation if provided. Defaults to ``True``.
        channel_info: additional information of this channel. Defaults to ``None``.

    Raises:
        ValueError: Unsupported channel representation for ``type_repr``.
        NotImplementedError: The noisy channel can only run in density matrix mode.
        TypeError: Unexpected data type for Channel representation.
        
    Note:
        Assume the input dimension is the same as the output.
    
    """
    def __init__(
        self, type_repr: str, representation: Optional[Union[torch.Tensor]] = None,
        system_idx: Optional[Union[Iterable[Iterable[int]], Iterable[int], int]] = None,
        acted_system_dim: Optional[Union[List[int], int]] = None, check_legality: Optional[bool] = True,
        channel_info: dict = None
    ) -> None:
        type_repr = type_repr.lower()
        if type_repr not in ['choi', 'kraus', 'gate', 'stinespring']:
            raise ValueError(
                "Unsupported channel representation:"
                f"require 'choi', 'kraus', 'gate', or 'stinespring', not {type_repr}")
        self.type_repr = type_repr
        super().__init__()
        self._info.update(channel_info or {})
        
        if representation is None:
            assert not isinstance(acted_system_dim, int), \
                f"Need to specify all system dimensions to get # of acted systems: received {acted_system_dim}"
            self.system_dim, num_acted_system = acted_system_dim, len(acted_system_dim)
        else:
            representation = representation.to(dtype=self.dtype, device=self.device)
            num_acted_system = getattr(self, f'_Channel__{type_repr}_init')(representation, acted_system_dim, check_legality)
        
        if system_idx is None:
            self.system_idx = [list(range(num_acted_system))]
        else:
            self.system_idx = _format_operator_idx(system_idx, num_acted_system)
        self.num_acted_system = num_acted_system
    
    def __call__(self, state: State) -> State:
        return self.forward(state)
    
    def __register_dim(self, input_dim: int, acted_system_dim: Union[List[int], int, None]) -> int:
        if acted_system_dim is None:
            self.system_dim, num_acted_system = [input_dim], 1
        elif isinstance(acted_system_dim, int):
            num_acted_system = int(math.log(input_dim, acted_system_dim))
            
            assert input_dim == (acted_system_dim ** num_acted_system), \
                f"The input dimension of {self.type_repr} representation does not match with acted system dimension: received {acted_system_dim}."
            
            self.system_dim = [acted_system_dim] * num_acted_system
        else:
            assert input_dim == (expec_dim := math.prod(acted_system_dim)), \
                (f"The input dimension of {self.type_repr} representation should match with system dimension {acted_system_dim}:" + 
                 f"input {input_dim}, expect {expec_dim}.")
            self.system_dim, num_acted_system = acted_system_dim, len(acted_system_dim)
        return num_acted_system

    def __choi_init(self, choi_repr: torch.Tensor, 
                    acted_system_dim: Union[List[int], int, None], 
                    check_legality: bool) -> int:
        r"""Initialize channel for type_repr as ``'choi'``
        """
        self._info['type'] = 'channel'
        
        input_dim = math.isqrt(choi_repr.shape[-1])
        num_acted_system = self.__register_dim(input_dim, acted_system_dim)

        if check_legality and not torch.all(utils.check._is_choi(choi_repr)):
            warnings.warn(
                "The input data may not be a Choi representation of a channel.", UserWarning)
        
        self.__choi_repr = choi_repr
        return num_acted_system

    def __kraus_init(self, kraus_repr: torch.Tensor, 
                     acted_system_dim: Union[List[int], int, None], 
                     check_legality: bool) -> int:
        r"""Initialize channel for type_repr as ``'kraus'``
        """
        self._info['type'] = 'channel'
        
        if len(kraus_repr.shape) == 2:
            kraus_repr = kraus_repr.unsqueeze(0)

        input_dim = kraus_repr.shape[-1]
        num_acted_system = self.__register_dim(input_dim, acted_system_dim)

        # sanity check
        if check_legality:
            oper_sum = (utils.linalg._dagger(kraus_repr) @ kraus_repr).sum(dim=-3)
            identity = torch.eye(input_dim).to(device=oper_sum.device)
            err = torch.norm(torch.abs(oper_sum - identity)).item()
            if err > min(1e-6 * input_dim * len(kraus_repr), 0.01):
                warnings.warn(
                    f"The input data may not be a Kraus representation of a channel: norm(sum(E * E^d) - I) = {err}.",
                    UserWarning)

        self.__kraus_repr = kraus_repr
        return num_acted_system
    
    def __gate_init(self, mat: torch.Tensor, 
                    acted_system_dim: Union[List[int], int, None], 
                    check_legality: bool) -> int:
        r"""Initialize channel for type_repr as ``'gate'``
        """
        self._info['type'] = 'gate'
        
        input_dim = mat.shape[-1]
        num_acted_system = self.__register_dim(input_dim, acted_system_dim)

        if check_legality:
            identity = torch.eye(input_dim).to(device=mat.device).expand_as(mat)
            err = torch.norm(torch.abs(utils.linalg._dagger(mat) @ mat - identity)).item()
            if err > min(1e-5 * input_dim, 0.01):
                warnings.warn(
                    f"\nThe input gate matrix may not be a unitary: norm(U * U^d - I) = {err}.", UserWarning)
        
        self.__kraus_repr = mat.unsqueeze(-3)
        return num_acted_system

    def __stinespring_init(self, stinespring_repr: torch.Tensor, 
                           acted_system_dim: Union[List[int], int, None], 
                           check_legality: bool) -> int:
        r"""Initialize channel for type_repr as ``'Stinespring'``
        
        """
        self._info['type'] = 'channel'
        
        input_dim = stinespring_repr.shape[-1]
        num_acted_system = self.__register_dim(input_dim, acted_system_dim)

        # sanity check
        if check_legality:
            # TODO: need to add more sanity check for stinespring
            out_dim = stinespring_repr.shape[-2]
            assert out_dim % input_dim == 0, \
                'The width of stinespring matrix should be the factor of its height'

        self.__stinespring_repr = stinespring_repr
        return num_acted_system

    @property
    def dim(self) -> int:
        r"""Dimension of the input/output system of this channel
        """
        return math.prod(self.system_dim)

    @property
    def choi_repr(self) -> torch.Tensor:
        r"""Choi representation of a channel

        Returns:
            a tensor with shape :math:`[d_\text{out}^2, d_\text{in}^2]`, where :math:`d_\text{in/out}` is
            the input/output dimension of this channel

        Raises:
            ValueError: Need to specify the Choi representation in this Channel instance.
        """
        type_repr = self.type_repr
        if type_repr == 'choi':
            if self.__choi_repr is None:
                raise ValueError(
                    "Need to specify the Choi representation in this Channel instance.")
            return self.__choi_repr
        elif type_repr == 'kraus':
            return _kraus_to_choi(self.kraus_repr)
        elif type_repr == 'gate':
            return _kraus_to_choi(self.matrix.unsqueeze(-3))
        else:
            return _stinespring_to_choi(self.stinespring_repr)

    @property
    def kraus_repr(self) -> List[torch.Tensor]:
        r"""Kraus representation of a channel

        Returns:
            a list of tensors with shape :math:`[d_\text{out}, d_\text{in}]`, where :math:`d_\text{in/out}` is
            the input/output dimension of this channel

        Raises:
            ValueError: Need to specify the Kraus representation in this Channel instance.
        """
        type_repr = self.type_repr

        if type_repr == 'choi':
            return _choi_to_kraus(self.choi_repr, tol=1e-6)
        elif type_repr == 'gate':
            return self.matrix.unsqueeze(-3)
        elif type_repr == 'kraus':
            if self.__kraus_repr is None:
                raise ValueError(
                    "Need to specify the Kraus representation in this Channel instance.")
            return self.__kraus_repr
        else:
            return _stinespring_to_kraus(self.stinespring_repr)

    @property
    def stinespring_repr(self) -> torch.Tensor:
        r"""Stinespring representation of a channel

        Returns:
            a tensor with shape :math:`[r * d_\text{out}, d_\text{in}]`, where :math:`r` is the rank of this channel and
            :math:`d_\text{in/out}` is the input/output dimension of this channel

        Raises:
            ValueError: Need to specify the Stinespring representation in this Channel instance.
        """
        type_repr = self.type_repr

        if type_repr == 'choi':
            return _choi_to_stinespring(self.choi_repr, tol=1e-6)
        elif type_repr == 'stinespring':
            if self.__stinespring_repr is None:
                raise ValueError(
                    "Need to specify the Stinespring representation in this Channel instance.")
            return self.__stinespring_repr
        else:
            return _kraus_to_stinespring(self.kraus_repr)
        
    @property
    def info(self) -> OperatorInfoType:
        r"""Information of this channel
        """
        info = super().info
        info.update({
            'system_idx': self.system_idx,
        })
        return info

    def forward(self, state: State) -> State:
        if state.backend == 'state_vector':
            state = state.to_mixed() 
        
        for system_idx in self.system_idx:
            if self.type_repr == 'choi':
                state._transform(self.choi_repr, system_idx, 'choi')
            else:
                state._transform(self.kraus_repr, system_idx, 'kraus')

        return state
